# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from functools import partial

import qtawesome as qta


BUTTON_ICON_COLOR = "#0081db"
TEXT_GREY = "#4d5157"
TEXT_LIGHTGREY = "#888f99"
WARNING_YELLOW = "#ff9e00"
ERROR_RED = "#e6635a"


# "partial" is used here as the icons from qtawesome needs a running QApplication
# in order to be constructed. This makes all of the "icons" below callable.
#
# Typical usage is (notice that it is a call, since "()" is present)
#
#     QLabel(ARROW_LEFT_BOLD(), ...)
#
# A nice side-effect is that the "color" of the icon can be overridden by the caller
#
#     QLabel(ARROW_LEFT_BOLD(color="<my custom color>"), ...)
#
# fmt: off
ARROW_LEFT_BOLD = partial(qta.icon, "ph.arrow-left-bold",        color=BUTTON_ICON_COLOR)
CALIBRATE       = partial(qta.icon, "fa5s.circle",               color=BUTTON_ICON_COLOR)
CHART_BAR       = partial(qta.icon, "fa5.chart-bar",             color=BUTTON_ICON_COLOR)
CHART_LINE      = partial(qta.icon, "ri.line-chart-fill",        color=BUTTON_ICON_COLOR)
CHECKMARK       = partial(qta.icon, "fa5s.check",                color=BUTTON_ICON_COLOR)
CLOSE           = partial(qta.icon, "mdi.close",                 color=BUTTON_ICON_COLOR)
COG             = partial(qta.icon, "fa5s.cog",                  color=BUTTON_ICON_COLOR)
EDIT            = partial(qta.icon, "fa5s.edit",                 color=BUTTON_ICON_COLOR)
EXTERNAL_LINK   = partial(qta.icon, "fa5s.external-link-alt",    color=BUTTON_ICON_COLOR)
FLASH           = partial(qta.icon, "mdi.flash",                 color=BUTTON_ICON_COLOR)
FOLDER_OPEN     = partial(qta.icon, "fa5s.folder-open",          color=BUTTON_ICON_COLOR)
GAUGE           = partial(qta.icon, "mdi.gauge-full",            color=BUTTON_ICON_COLOR)
HELP            = partial(qta.icon, "mdi6.help-circle",          color=BUTTON_ICON_COLOR)
INFO            = partial(qta.icon, "fa5s.info-circle",          color=BUTTON_ICON_COLOR)
LINK            = partial(qta.icon, "fa5s.link",                 color=BUTTON_ICON_COLOR)
MEMORY          = partial(qta.icon, "fa5s.memory",               color=BUTTON_ICON_COLOR)
PLAY            = partial(qta.icon, "fa5s.play-circle",          color=BUTTON_ICON_COLOR)
PLUS            = partial(qta.icon, "ei.plus-sign",              color=BUTTON_ICON_COLOR)
REFRESH         = partial(qta.icon, "ei.refresh",                color=BUTTON_ICON_COLOR)
REMOVE          = partial(qta.icon, "ei.remove-sign",            color=BUTTON_ICON_COLOR)
RESTORE         = partial(qta.icon, "mdi6.restore",              color=BUTTON_ICON_COLOR)
SAVE            = partial(qta.icon, "mdi.content-save",          color=BUTTON_ICON_COLOR)
RECORD          = partial(qta.icon, "mdi.record-circle-outline", color=BUTTON_ICON_COLOR)
STOP            = partial(qta.icon, "fa5s.stop-circle",          color=BUTTON_ICON_COLOR)
UNLINK          = partial(qta.icon, "fa5s.unlink",               color=BUTTON_ICON_COLOR)
WARNING         = partial(qta.icon, "ph.warning-fill",           color=WARNING_YELLOW)
EYE_OPEN        = partial(qta.icon, "fa5s.eye",                  color=BUTTON_ICON_COLOR)
EYE_CLOSED      = partial(qta.icon, "fa5s.eye-slash",            color=BUTTON_ICON_COLOR)
# fmt: on



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
from __future__ import annotations

import typing as t

import attrs
import typing_extensions as te

from acconeer.exptool import a121
from acconeer.exptool.a121._core import utils as core_utils

from . import domain
from .lookup import Module, Sensor


_S = t.TypeVar("_S")
_T = t.TypeVar("_T")
_U = t.TypeVar("_U")
_ms = _mA = 1e-3
_us = _uA = 1e-6


@te.overload
def _interleave(
    __a: t.Iterable[_S],
    __b: t.Iterable[_T],
    __c: t.Iterable[_U],
) -> t.Iterator[t.Union[_S, _T, _U]]: ...


@te.overload
def _interleave(
    __a: t.Iterable[_S],
    __b: t.Iterable[_T],
) -> t.Iterator[t.Union[_S, _T]]: ...


def _interleave(*iterables: t.Iterable[_T]) -> t.Iterator[_T]:
    iterators = [iter(i) for i in iterables]

    while True:
        for iterator in iterators:
            try:
                yield next(iterator)
            except StopIteration:
                return


class Algorithm(te.Protocol):
    def decide_control(
        self,
        frame_actives: t.Sequence[t.Mapping[int, domain.EnergyRegion]],
        frame_idles: list[dict[int, domain.SimpleRegion]],
        config: a121.SessionConfig,
        lower_idle_state: t.Optional[Sensor.LowerIdleState],
        sensor: Sensor,
        module: Module,
    ) -> t.Iterable[domain.EnergyRegion]: ...


class SparseIq(Algorithm):
    def decide_control(
        self,
        frame_actives: t.Sequence[t.Mapping[int, domain.EnergyRegion]],
        frame_idles: list[dict[int, domain.SimpleRegion]],
        config: a121.SessionConfig,
        lower_idle_state: t.Optional[Sensor.LowerIdleState],
        sensor: Sensor,
        module: Module,
    ) -> t.Iterable[domain.EnergyRegion]:
        active_iter = core_utils.iterate_extended_structure_values(frame_actives)
        idle_list = list(core_utils.iterate_extended_structure_values(frame_idles))[:-1]

        if lower_idle_state == Sensor.IdleState.OFF:
            yield off_exit(sensor, module)
        elif lower_idle_state == Sensor.IdleState.HIBERNATE:
            yield hibernate_exit(last_inter_frame_idle_state(config), sensor, module)

        yield from _interleave(
            active_iter,
            idle_list,
        )

        if lower_idle_state == Sensor.IdleState.OFF:
            yield off_enter(sensor, module)
        elif lower_idle_state == Sensor.IdleState.HIBERNATE:
            yield hibernate_enter(sensor, module)


class Presence(Algorithm):
    _PROCESSING_SECONDS_PER_POINT = 11.28125e-6

    def decide_control(
        self,
        frame_actives: t.Sequence[t.Mapping[int, domain.EnergyRegion]],
        frame_idles: list[dict[int, domain.SimpleRegion]],
        config: a121.SessionConfig,
        lower_idle_state: t.Optional[Sensor.LowerIdleState],
        sensor: Sensor,
        module: Module,
    ) -> t.Iterable[domain.EnergyRegion]:
        (sensor_config,) = config.groups[0].values()
        num_points = sum([subsweep.num_points for subsweep in sensor_config.subsweeps])
        points_measured = num_points * sensor_config.sweeps_per_frame
        process_duration = points_measured * self._PROCESSING_SECONDS_PER_POINT

        yield from SparseIq().decide_control(
            frame_actives, frame_idles, config, lower_idle_state, sensor, module
        )
        yield domain.SimpleRegion(
            module.currents[Module.PowerState.PROCESSING],
            process_duration,
            tag=domain.EnergyRegion.Tag.PROCESSING,
            description="Presence processing",
        )


class Distance(Algorithm):
    _PROCESSING_SECONDS_PER_POINT = 12e-6

    @classmethod
    def processing_region(
        cls,
        sensor_config: a121.SensorConfig,
        sensor: Sensor,
        module: Module,
    ) -> domain.SimpleRegion:
        static_duration = 2.5e-3

        num_points = sum(ssc.num_points for ssc in sensor_config.subsweeps)
        return domain.SimpleRegion(
            (
                module.currents[Module.PowerState.PROCESSING]
                + sensor.inter_frame_currents[sensor_config.inter_frame_idle_state]
            ),
            duration=static_duration + num_points * cls._PROCESSING_SECONDS_PER_POINT,
            tag=domain.EnergyRegion.Tag.PROCESSING,
            description=f"Distance processing ({num_points} points)",
        )

    @staticmethod
    def recalculate_idle(
        idle_and_processing: tuple[domain.SimpleRegion, domain.EnergyRegion],
    ) -> t.Optional[domain.SimpleRegion]:
        idle, processing = idle_and_processing

        if idle.duration < processing.duration:
            return None
        else:
            return attrs.evolve(idle, duration=idle.duration - processing.duration)

    def decide_control(
        self,
        frame_actives: t.Sequence[t.Mapping[int, domain.EnergyRegion]],
        frame_idles: list[dict[int, domain.SimpleRegion]],
        config: a121.SessionConfig,
        lower_idle_state: t.Optional[Sensor.LowerIdleState],
        sensor: Sensor,
        module: Module,
    ) -> t.Iterator[domain.EnergyRegion]:
        processings = core_utils.map_over_extended_structure(
            lambda sc: self.processing_region(sc, sensor=sensor, module=module),
            config.groups,
        )

        idles = core_utils.map_over_extended_structure(
            self.recalculate_idle,
            core_utils.zip_extended_structures(
                frame_idles,
                processings,
            ),
        )

        if lower_idle_state == Sensor.IdleState.OFF:
            yield off_exit(sensor, module)
        elif lower_idle_state == Sensor.IdleState.HIBERNATE:
            yield hibernate_exit(last_inter_frame_idle_state(config), sensor, module)

        for region in _interleave(
            core_utils.iterate_extended_structure_values(frame_actives),
            core_utils.iterate_extended_structure_values(processings),
            core_utils.iterate_extended_structure_values(idles),
        ):
            if region is not None:
                yield region

        if lower_idle_state == Sensor.IdleState.OFF:
            yield off_enter(sensor, module)
        elif lower_idle_state == Sensor.IdleState.HIBERNATE:
            yield hibernate_enter(sensor, module)


def hibernate_enter(
    sensor: Sensor,
    module: Module,
) -> domain.SimpleRegion:
    """
    Returns a region describing the sequence of hibernating
    """
    set_hibernate_register = domain.SimpleRegion(
        module.currents[Module.PowerState.IO],
        110 * _us,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Tell sensor to go into hibernation",
    )
    wfi = domain.SimpleRegion(
        module.currents[Module.PowerState.AWAKE],
        2 * _ms,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Wait for interrupt",
    )

    real_enter = domain.CompositeRegion(
        (set_hibernate_register, wfi),
        description="Go down to hibernation",
    )

    return domain.SimpleRegion(
        real_enter.average_current,
        real_enter.duration,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Go down to hibernation",
    )


def hibernate_exit(
    recent_inter_frame_idle: a121.IdleState,
    sensor: Sensor,
    module: Module,
) -> domain.SimpleRegion:
    """
    Returns a region describing the sequence of waking up from hibernation
    """
    start_interfaces = domain.SimpleRegion(
        module.currents[Module.PowerState.IO],
        0.1 * _ms,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Starting interfaces",
    )
    wfi = domain.SimpleRegion(
        module.currents[Module.PowerState.AWAKE]
        + sensor.inter_frame_currents[recent_inter_frame_idle],
        2 * _ms,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Wait for interrupt",
    )
    spi_sanity = domain.SimpleRegion(
        module.currents[Module.PowerState.IO]
        + sensor.inter_frame_currents[recent_inter_frame_idle],
        300 * _us,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Spi sanity",
    )

    real_exit = domain.CompositeRegion(
        (start_interfaces, wfi, spi_sanity), description="Wake up from hibernate"
    )

    return domain.SimpleRegion(
        real_exit.average_current,
        real_exit.duration,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Wake up from hibernate",
    )


def off_enter(
    sensor: Sensor,
    module: Module,
) -> domain.SimpleRegion:
    """
    Describes the power region of powering down (entering off)
    """
    wfi = domain.SimpleRegion(
        module.currents[Module.PowerState.AWAKE],
        2 * _ms,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Enter off",
    )

    return wfi


def off_exit(
    sensor: Sensor,
    module: Module,
) -> domain.SimpleRegion:
    """
    Returns a region describing the sequence of waking up from off
    """

    return domain.SimpleRegion(
        10.51 * _mA,
        4 * _ms,
        tag=domain.EnergyRegion.Tag.COMMUNICATION,
        description="Wake up from off",
    )


def last_inter_frame_idle_state(config: a121.SessionConfig) -> a121.IdleState:
    last_sensor_config = list(config.groups[-1].values())[-1]
    return last_sensor_config.inter_frame_idle_state



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import numpy as np

from acconeer.exptool.a121 import SensorConfig, SessionConfig
from acconeer.exptool.a121._core import utils as core_utils
from acconeer.exptool.a121.algo import distance, get_distance_filter_edge_margin
from acconeer.exptool.a121.algo.distance import (
    Detector as DistanceDetector,
)
from acconeer.exptool.a121.algo.distance import (
    DetectorConfig as DistanceConfig,
)
from acconeer.exptool.a121.algo.distance import (
    MeasurementType,
    ThresholdMethod,
)
from acconeer.exptool.a121.algo.distance._utils import (
    get_calibrate_noise_session_config,
    get_calibrate_offset_sensor_config,
)
from acconeer.exptool.a121.algo.presence import (
    Detector as PresenceDetector,
)
from acconeer.exptool.a121.algo.presence import (
    DetectorConfig as PresenceConfig,
)


OVERHEAD = 68
CALIB_BUFFER = 2492
BYTES_PER_POINT = 4
MAX_NUM_POINTS = 4095

RSS_HEAP_PER_SUBSWEEP = 236
RSS_HEAP_PER_SENSOR = 636
RSS_HEAP_PER_CONFIG = 512

SIZE_OF_FLOAT = 4

PRESENCE_HEAP_OVERHEAD = 256

DISTANCE_HEAP_OVERHEAD = 1028
DISTANCE_HEAP_PER_PROCESSOR = 224


def _sweep_external_heap_memory(config: SensorConfig) -> int:
    total_num_points = (
        np.sum([subconfig.num_points for subconfig in config.subsweeps]) * config.sweeps_per_frame
    )

    total_bytes = max(total_num_points * BYTES_PER_POINT, CALIB_BUFFER) + OVERHEAD

    return int(total_bytes)


def _sweep_rss_heap_memory(config: SensorConfig) -> int:
    return RSS_HEAP_PER_CONFIG + config.num_subsweeps * RSS_HEAP_PER_SUBSWEEP


def session_external_heap_memory(config: SessionConfig) -> int:
    sensor_cfgs = list(core_utils.iterate_extended_structure_values(config.groups))

    return max([_sweep_external_heap_memory(cfg) for cfg in sensor_cfgs])


def _session_config_rss_heap_memory(config: SessionConfig) -> int:
    sensor_cfgs = list(core_utils.iterate_extended_structure_values(config.groups))

    return sum([_sweep_rss_heap_memory(cfg) for cfg in sensor_cfgs])


def session_rss_heap_memory(config: SessionConfig) -> int:
    sensor_ids = set([x[1] for x in core_utils.iterate_extended_structure(config.groups)])

    return RSS_HEAP_PER_SENSOR * len(sensor_ids) + _session_config_rss_heap_memory(config)


def session_heap_memory(config: SessionConfig) -> int:
    return session_external_heap_memory(config) + session_rss_heap_memory(config)


def presence_external_heap_memory(config: PresenceConfig) -> int:
    sensor_cfg = PresenceDetector._get_sensor_config(config)

    session_ext_heap = session_external_heap_memory(SessionConfig(sensor_cfg))
    num_points = sum([subsweep.num_points for subsweep in sensor_cfg.subsweeps])
    prec_ext_heap = num_points * 2 * SIZE_OF_FLOAT

    return session_ext_heap + prec_ext_heap


def presence_rss_heap_memory(config: PresenceConfig) -> int:
    sensor_cfg = PresenceDetector._get_sensor_config(config)

    PREC_FILTER_PARAMS = 7

    num_points = sum([subsweep.num_points for subsweep in sensor_cfg.subsweeps])
    prec_rss_heap = num_points * PREC_FILTER_PARAMS * SIZE_OF_FLOAT

    return (
        PRESENCE_HEAP_OVERHEAD
        + prec_rss_heap
        + RSS_HEAP_PER_SENSOR
        + _sweep_rss_heap_memory(sensor_cfg)
    )


def presence_heap_memory(config: PresenceConfig) -> int:
    return presence_external_heap_memory(config) + presence_rss_heap_memory(config)


def distance_external_heap_memory(config: DistanceConfig) -> int:
    offset_sensor_config = get_calibrate_offset_sensor_config()
    session_config = distance.detector_config_to_session_config(config, [1])
    processor_specs = distance.detector_config_to_processor_specs(config, [1], 4)
    noise_session_config = get_calibrate_noise_session_config(session_config, [1])

    offset_ext_heap = session_external_heap_memory(SessionConfig(offset_sensor_config))
    session_ext_heap = session_external_heap_memory(session_config)
    noise_ext_heap = session_external_heap_memory(noise_session_config)

    sensor_buffer = max([offset_ext_heap, session_ext_heap, noise_ext_heap])

    NOISE_NUM_POINTS = 220
    FILTFILT_PAD_LEN = 9

    noise_work_buffer = (NOISE_NUM_POINTS + 2 * FILTFILT_PAD_LEN) * 2 * SIZE_OF_FLOAT

    sensor_cfgs = list(core_utils.iterate_extended_structure_values(session_config.groups))

    aggr_work_buffer = 0
    det_calib_buffer = 2 * SIZE_OF_FLOAT
    aggr_calib_buffer = 0

    for proc_spec in processor_specs:
        proc_work_buffer = 0
        proc_calib_buffer = 0
        sensor_cfg = sensor_cfgs[proc_spec.group_index]

        if proc_spec.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            subsweep_cfgs = [sensor_cfg.subsweeps[1]]
            det_calib_buffer += SIZE_OF_FLOAT
        else:
            subsweep_cfgs = [sensor_cfg.subsweeps[i] for i in proc_spec.subsweep_indexes]
            det_calib_buffer += len(subsweep_cfgs) * SIZE_OF_FLOAT

        num_points = sum([sub_cfg.num_points for sub_cfg in subsweep_cfgs])
        filt_margin = get_distance_filter_edge_margin(
            subsweep_cfgs[0].profile, subsweep_cfgs[0].step_length
        )
        num_points_cropped = num_points - 2 * filt_margin
        half_num_points_cropped = num_points_cropped // 2
        aligned_half_num_points_cropped = (
            half_num_points_cropped
            if half_num_points_cropped % 2 == 0
            else half_num_points_cropped + 1
        )

        proc_work_buffer = (num_points + 2 * FILTFILT_PAD_LEN) * 2 * SIZE_OF_FLOAT
        proc_work_buffer += ((num_points_cropped + (32 - 1)) // 32) * SIZE_OF_FLOAT
        proc_work_buffer += aligned_half_num_points_cropped * 2

        if proc_spec.processor_config.threshold_method == ThresholdMethod.RECORDED:
            proc_calib_buffer += num_points_cropped * SIZE_OF_FLOAT * 3
            proc_calib_buffer += len(subsweep_cfgs) * SIZE_OF_FLOAT

        if proc_spec.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            proc_calib_buffer += sensor_cfg.sweeps_per_frame * num_points * SIZE_OF_FLOAT
            proc_calib_buffer += sensor_cfg.sweeps_per_frame * SIZE_OF_FLOAT

        aggr_work_buffer += proc_work_buffer
        aggr_calib_buffer += proc_calib_buffer

    work_buffer = max([noise_work_buffer, aggr_work_buffer])

    return sensor_buffer + work_buffer + det_calib_buffer + aggr_calib_buffer


def distance_rss_heap_memory(config: DistanceConfig) -> int:
    offset_sensor_config = get_calibrate_offset_sensor_config()
    session_config = distance.detector_config_to_session_config(config, [1])
    processor_specs = distance.detector_config_to_processor_specs(config, [1], 4)
    noise_session_config = get_calibrate_noise_session_config(session_config, [1])

    offset_rss_heap = _session_config_rss_heap_memory(SessionConfig(offset_sensor_config))
    session_rss_heap = _session_config_rss_heap_memory(session_config)
    noise_rss_heap = _session_config_rss_heap_memory(noise_session_config)

    # Loopback sweep is not part of noise calibration
    if DistanceDetector._has_close_range_measurement(config):
        noise_rss_heap = noise_rss_heap - RSS_HEAP_PER_SUBSWEEP

    sensor_heap = RSS_HEAP_PER_SENSOR

    processor_heap = DISTANCE_HEAP_PER_PROCESSOR * len(processor_specs)

    return (
        DISTANCE_HEAP_OVERHEAD
        + processor_heap
        + sensor_heap
        + offset_rss_heap
        + session_rss_heap
        + noise_rss_heap
    )


def distance_heap_memory(config: DistanceConfig) -> int:
    return distance_external_heap_memory(config) + distance_rss_heap_memory(config)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

from typing import Optional


class DeviceFlasherBase:
    def flash(self, device_port, device_name, image_path, progress_callback=None) -> None:
        pass

    @staticmethod
    def get_boot_description(device_name: str) -> Optional[str]:
        return None

    @staticmethod
    def get_post_dfu_description(device_name: str) -> Optional[str]:
        return None



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import itertools
import time
import warnings
from typing import Any, Iterator, Optional, Union

import acconeer.exptool.a121._core.utils as core_utils
from acconeer.exptool._core import ClientInfo
from acconeer.exptool._core.communication.client import ClientError
from acconeer.exptool.a121 import (
    Client,
    Metadata,
    Record,
    Result,
    SensorCalibration,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)


class _StopReplay(Exception):
    pass


class ReplaySessionsExhaustedError(Exception):
    pass


class _ReplayingClient(Client, register=False):
    def __init__(
        self,
        record: Record,
        *,
        cycled_session_idx: Optional[int] = None,
        realtime_replay: bool = True,
    ):
        """
        :param record: The Record to replay from
        :param cycled_session_idx:
            If specified, cycle (reuse as next session) the session
            specified by 'cycled_session_idx'.
        :param realtime_replay: If True, replays the data at the rate of recording
        """
        super().__init__(record.client_info)
        self._record = record
        self._is_started: bool = False
        self._result_iterator: Iterator[list[dict[int, Result]]] = iter([])
        self._origin_time: Optional[float] = None
        self._realtime_replay = realtime_replay
        self._session_idx: Optional[int] = None
        self._session_idx_iter: Union[Iterator[int], itertools.repeat[int]]
        if cycled_session_idx is not None:
            self._session_idx_iter = itertools.repeat(cycled_session_idx)
        else:
            self._session_idx_iter = iter(range(record.num_sessions))

    @property
    def _actual_session_idx(self) -> int:
        if self._session_idx is not None:
            return self._session_idx
        else:
            msg = "Session is not set up."
            raise ClientError(msg)

    def _assert_connected(self) -> None:
        if not self.connected:
            msg = "Client is not connected."
            raise ClientError(msg)

    def _assert_session_setup(self) -> None:
        self._assert_connected()
        if not self.session_is_setup:
            msg = "Session is not set up."
            raise ClientError(msg)

    def _assert_session_started(self) -> None:
        self._assert_session_setup()
        if not self.session_is_started:
            msg = "Session is not started."
            raise ClientError(msg)

    def _open(self) -> None:
        pass

    def setup_session(  # type: ignore[override]
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        if isinstance(config, SensorConfig):
            config = SessionConfig(config)

        try:
            new_session_idx = next(self._session_idx_iter)
        except StopIteration:
            raise ReplaySessionsExhaustedError
        else:
            self._session_idx = new_session_idx

        if config != self.session_config:
            raise ValueError

        if self.session_config.extended:
            return self.extended_metadata
        else:
            return core_utils.unextend(self.extended_metadata)

    def start_session(self) -> None:
        self._result_iterator = self._record.session(self._actual_session_idx).extended_results
        self._is_started = True
        self._origin_time = None

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:  # type: ignore[override]
        if not self.session_is_setup:
            msg = "Session is not set up."
            raise ClientError(msg)

        assert self._result_iterator is not None

        try:
            result = next(self._result_iterator)
        except StopIteration:
            raise _StopReplay

        some_result = next(core_utils.iterate_extended_structure_values(result))

        if self._realtime_replay:
            now = time.monotonic() - some_result.tick_time

            if self._origin_time is None:
                self._origin_time = now

            delta = now - self._origin_time

            if delta < 0:
                time.sleep(-delta)

        if self.session_config.extended:
            return result
        else:
            return core_utils.unextend(result)

    def stop_session(self) -> Any:
        try:
            _ = next(self._result_iterator)
        except StopIteration:
            pass
        else:
            warnings.warn(f"Results of session {self._actual_session_idx} were not exhausted.")

        self._is_started = False

    def close(self) -> None:
        pass

    @property
    def connected(self) -> bool:
        return True

    @property
    def session_is_setup(self) -> bool:
        return True

    @property
    def session_is_started(self) -> bool:
        return self._is_started

    @property
    def server_info(self) -> ServerInfo:
        return self._record.server_info

    @property
    def client_info(self) -> ClientInfo:
        return self._record.client_info

    @property
    def session_config(self) -> SessionConfig:
        return self._record.session(self._actual_session_idx).session_config

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        return self._record.session(self._actual_session_idx).extended_metadata

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        return self._record.session(self._actual_session_idx).calibrations

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        return self._record.session(self._actual_session_idx).calibrations_provided



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import typing as t

from acconeer.exptool.app.new import AppModel


class NullAppModel(AppModel):
    class _NullSignal:
        def connect(self, slot: t.Callable[..., t.Any]) -> None:
            pass

    sig_notify: _NullSignal  # type: ignore[assignment]
    sig_message_plot_plugin: _NullSignal  # type: ignore[assignment]

    def __init__(self) -> None:
        self.sig_notify = self._NullSignal()
        self.sig_message_plot_plugin = self._NullSignal()
        self.sig_backend_state_changed = self._NullSignal()  # type: ignore[assignment]
        self.sig_load_plugin = self._NullSignal()  # type: ignore[assignment]



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

from typing import Optional, Union

from PySide6 import QtCore
from PySide6.QtWidgets import (
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QStackedLayout,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool import a121
from acconeer.exptool.app.new.ui.components.group_box import GroupBox


_WIDGET_WIDTH = 125


class SmartPerfCalcView(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self._layout = QStackedLayout()
        self.setLayout(self._layout)

        self._perf_calc_view = PerfCalcView(self)
        self._extended_perf_calc_view = ExtendedPerfCalcView(self)
        self._layout.addWidget(self._perf_calc_view)
        self._layout.addWidget(self._extended_perf_calc_view)

    def set_data(
        self,
        session_config: Optional[a121.SessionConfig] = None,
        metadata: Optional[Union[a121.Metadata, list[dict[int, a121.Metadata]]]] = None,
    ) -> None:
        if isinstance(metadata, list):
            self._layout.setCurrentIndex(1)
            self._extended_perf_calc_view.set_data(session_config, metadata)
        else:
            self._layout.setCurrentIndex(0)
            self._perf_calc_view.set_data(session_config, metadata)


class ExtendedPerfCalcView(GroupBox[QVBoxLayout]):
    def __init__(self, parent: QWidget) -> None:
        super().__init__("Performance calculations", QVBoxLayout, parent=parent)

        label = QLabel("No estimates available", self)
        label.setEnabled(False)
        label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.layout().addWidget(label)

    def set_data(
        self,
        session_config: Optional[a121.SessionConfig] = None,
        extended_metadata: Optional[list[dict[int, a121.Metadata]]] = None,
    ) -> None:
        pass


class PerfCalcView(QGroupBox):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setTitle("Performance estimates")

        tooltip = (
            "The estimate is based on calculations and "
            "conservative measurements in room temperature."
        )

        layout = QGridLayout(self)
        self.setLayout(layout)

        self.average_current = PerfCalcValueWidget(self)
        layout.addWidget(self.average_current, 0, 1)
        average_current_label = QLabel("Estimated avg. current:", self)
        average_current_label.setToolTip(tooltip)
        layout.addWidget(average_current_label, 0, 0)

        self.set_data(None)

    def set_data(
        self,
        session_config: Optional[a121.SessionConfig] = None,
        metadata: Optional[a121.Metadata] = None,
    ) -> None:
        try:
            self._update(session_config, metadata)
        except Exception:
            self.average_current.setText("-")

    def _update(
        self,
        session_config: Optional[a121.SessionConfig] = None,
        metadata: Optional[a121.Metadata] = None,
    ) -> None:
        if session_config is None or metadata is None:
            raise ValueError

        pc = a121._SensorPerformanceCalc(
            session_config.sensor_config, metadata, session_config.update_rate
        )

        self.average_current.setText(f"{pc.average_current * 1e3:.0f} mA")


class PerfCalcValueWidget(QLineEdit):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self.setFixedWidth(_WIDGET_WIDTH)
        self.setReadOnly(True)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from __future__ import annotations

import abc
import json
import re
from typing import Any, Iterator, List, Optional

import attrs
import serial.tools
import serial.tools.list_ports
import typing_extensions as te
import usb.core
from packaging import version

from .links.usb_link import get_libusb_backend


class CommDeviceError(Exception):
    pass


_USB_IDS = [  # (vid, pid, 'model number', 'Unflashed')
    (0x0483, 0xA41D, "XC120", True),
    (0x0483, 0xA42C, "XC120", True),
    (0x0483, 0xA42D, "XC120", False),
    (0x0483, 0xA449, "XC120", False),
    (0xACC0, 0xE121, "XV12X", False),
]


@attrs.frozen(kw_only=True)
class CommDevice(abc.ABC):
    name: Optional[str] = None
    serial: Optional[str] = None
    unflashed: bool = False
    recognized: bool = False

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> te.Self:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> te.Self:
        return cls.from_dict(json.loads(json_str))

    @abc.abstractmethod
    def display_name(self) -> str:
        pass


@attrs.frozen(kw_only=True)
class SerialDevice(CommDevice):
    port: str

    def display_name(self) -> str:
        if self.name is not None:
            dev_name = f"{self.name} {self.port}"
        else:
            dev_name = f"{self.port}"

        if self.unflashed:
            display_name = f"Unflashed {dev_name}"
        elif self.serial is not None:
            display_name = f"{dev_name} ({self.serial})"
        else:
            display_name = dev_name
        return display_name

    def __str__(self) -> str:
        name_str = "" if self.name is None else f"{self.name}: "
        serial_str = "?" if self.serial is None else self.serial
        return f"{name_str}port={self.port} serial={serial_str}"


@attrs.frozen(kw_only=True)
class USBDevice(CommDevice):
    vid: int
    pid: int
    accessible: bool = True

    def display_name(self) -> str:
        if not self.accessible:
            display_name = f"{self.name} (inaccessible)"
        elif self.unflashed:
            display_name = f"Unflashed {self.name}"
        elif self.serial is not None:
            display_name = f"{self.name} ({self.serial})"
        else:
            display_name = f"{self.name}"
        return display_name

    def __str__(self) -> str:
        serial_str = "?" if self.serial is None else self.serial
        return (
            f"{self.name}: USB_VID=0x{self.vid:04x}, USB_PID=0x{self.pid:04x} serial={serial_str}"
        )


def serial_device_from_port_object(port_object: serial.Serial) -> SerialDevice:
    # Return USB serial port if existing in _USB_IDS
    for vid, pid, model_number, unflashed in _USB_IDS:
        if port_object.vid == vid and port_object.pid == pid:
            return SerialDevice(
                name=model_number,
                port=port_object.device,
                serial=port_object.serial_number,
                unflashed=unflashed,
                recognized=True,
            )

    # Find serial portTryReturn USB serial port if existing in _USB_IDS
    device_port = port_object.device
    device_name = None
    device_serial_number = port_object.serial_number
    device_unflashed = False
    device_recognized = False

    PRODUCT_REGEX = r"[X][A-Z]\d{3}"
    desc = port_object.product or port_object.description
    match = re.search(PRODUCT_REGEX, desc)

    if match is None:
        pass
    elif match.group().lower() in ["xe123", "xe124", "xe125", "xe132"]:
        if version.parse(serial.__version__) >= version.parse("3.5"):
            # Special handling of cp2105 modules with with pyserial >= 3.5
            interface = port_object.interface

            if interface and "enhanced" in interface.lower():
                # Add the "enhanced" interface
                device_name = match.group().upper()
                device_recognized = True

        else:  # pyserial <= 3.4
            # Add "?" to both to indicate that it could be either.
            device_name = f"{match.group().upper()} (?)"
    else:
        device_recognized = True
        device_name = f"{match.group().upper()}"

    if desc is not None and "Bootloader" in desc:
        device_unflashed = True

    return SerialDevice(
        name=device_name,
        port=device_port,
        serial=device_serial_number,
        unflashed=device_unflashed,
        recognized=device_recognized,
    )


def get_serial_devices() -> List[SerialDevice]:
    NATIVE_SERIAL_PORT_FILTER = re.compile(r"^/dev/ttyS\d{1,2}$")
    serial_devices = []

    port_objects = serial.tools.list_ports.comports()
    for port_object in port_objects:
        if bool(NATIVE_SERIAL_PORT_FILTER.match(port_object.device)):
            continue
        serial_devices.append(serial_device_from_port_object(port_object))

    return serial_devices


class _UsbDeviceFinder:
    device_cache: dict[str, bool] = {}
    serial_port_cache: dict[str, bool] = {}

    def __init__(self) -> None:
        self._backend = get_libusb_backend()

    def iterate_devices(self) -> Iterator[tuple[int, int, Optional[str]]]:
        devices = usb.core.find(find_all=True, backend=self._backend)
        for dev in devices:
            try:
                serial_number = dev.serial_number
            except (ValueError, NotImplementedError):
                serial_number = None
            vid = dev.idVendor
            pid = dev.idProduct
            if self._is_com_port(vid, pid):
                continue
            usb.util.dispose_resources(dev)
            yield (vid, pid, serial_number)

    def _is_com_port(self, vid: int, pid: int) -> bool:
        vid_pid_str = f"{vid:04x}:{pid:04x}"
        if vid_pid_str not in self.serial_port_cache:
            port_objects = serial.tools.list_ports.comports()
            for port_object in port_objects:
                if port_object.vid is None or port_object.pid is None:
                    continue
                port_vid_pid_str = f"{port_object.vid:04x}:{port_object.pid:04x}"
                self.serial_port_cache[port_vid_pid_str] = True

        if vid_pid_str not in self.serial_port_cache:
            self.serial_port_cache[vid_pid_str] = False

        return self.serial_port_cache[vid_pid_str]

    def is_accessible(self, vid: int, pid: int) -> bool:
        vid_pid_str = f"{vid:04x}:{pid:04x}"
        if vid_pid_str not in self.device_cache:
            try:
                device = usb.core.find(idVendor=vid, idProduct=pid, backend=self._backend)
                # This will raise an USBError exception if inaccessible
                device.set_configuration()
                self.device_cache[vid_pid_str] = True
            except usb.core.USBError:
                self.device_cache[vid_pid_str] = False
            except NotImplementedError:
                # The function set_configuration is not implemented in windows libusb
                self.device_cache[vid_pid_str] = True

        return self.device_cache[vid_pid_str]


def get_usb_devices(only_accessible: bool = False) -> List[USBDevice]:
    usb_devices: List[USBDevice] = []

    usb_device_finder = _UsbDeviceFinder()
    for device_vid, device_pid, serial_number in usb_device_finder.iterate_devices():
        for vid, pid, model_name, unflashed in _USB_IDS:
            if device_vid == vid and device_pid == pid:
                device_name = model_name
                accessible = usb_device_finder.is_accessible(vid, pid)
                if only_accessible and not accessible:
                    continue

                usb_devices.append(
                    USBDevice(
                        vid=device_vid,
                        pid=device_pid,
                        serial=serial_number,
                        name=device_name,
                        accessible=accessible,
                        unflashed=unflashed,
                        recognized=True,
                    )
                )

    return usb_devices


def get_usb_device_by_serial(serial: str, only_accessible: bool = False) -> USBDevice:
    usb_devices = get_usb_devices(only_accessible=only_accessible)
    if serial is not None:
        for device in usb_devices:
            if serial == device.serial:
                return device
    msg = f"Could not find usb device with serial number '{serial}'"
    raise ValueError(msg)


def get_one_usb_device(only_accessible: bool = False) -> USBDevice:
    usb_devices = get_usb_devices(only_accessible=only_accessible)
    if not usb_devices:
        msg = "No USB devices detected. Cannot auto detect."
        raise CommDeviceError(msg)
    elif len(usb_devices) > 1:
        devices_string = "".join([f" - {dev}\n" for dev in usb_devices])
        raise CommDeviceError(
            "There are multiple devices detected. Specify one:\n" + devices_string
        )
    else:
        return usb_devices[0]


def get_one_serial_device() -> SerialDevice:
    acconeer_serial_devices: List[SerialDevice] = [
        device for device in get_serial_devices() if device.name is not None
    ]

    if not acconeer_serial_devices:
        msg = "No serial devices detected. Cannot auto detect."
        raise CommDeviceError(msg)
    elif len(acconeer_serial_devices) > 1:
        devices_string = "".join([f" - {dev}\n" for dev in acconeer_serial_devices])
        raise CommDeviceError(
            "There are multiple devices detected. Specify one:\n" + devices_string
        )
    else:
        return acconeer_serial_devices[0]


def tag_serial_ports_objects(
    port_infos: list[serial.Serial],
) -> list[tuple[serial.Serial, Optional[str]]]:
    PRODUCT_REGEX = r"[X][A-Z]\d{3}"

    port_tag_tuples: list[tuple[serial.Serial, Optional[str]]] = []

    for port_object in port_infos:
        desc = port_object.product or port_object.description

        match = re.search(PRODUCT_REGEX, desc)
        if match is None:
            for vid, pid, model_number, _ in _USB_IDS:
                if port_object.vid == vid and port_object.pid == pid:
                    port_tag_tuples.append((port_object, model_number))
                    break
            else:
                port_tag_tuples.append((port_object, None))
        elif match.group().lower() in ["xe123", "xe124", "xe125", "xe132"]:
            if version.parse(serial.__version__) >= version.parse("3.5"):
                # Special handling of cp2105 modules with with pyserial >= 3.5
                interface = port_object.interface

                if interface and "enhanced" in interface.lower():
                    # Add the "enhanced" interface
                    port_tag_tuples.append((port_object, f"{match.group().upper()}"))
                else:
                    # Add the "standard" interface but don't tag it
                    port_tag_tuples.append((port_object, None))

            else:  # pyserial <= 3.4
                # Add "?" to both to indicate that it could be either.
                port_tag_tuples.append((port_object, f"{match.group().upper()} (?)"))
        elif "Bootloader" in desc:
            port_tag_tuples.append((port_object, f"Unflashed {match.group()}"))
        else:
            port_tag_tuples.append((port_object, f"{match.group()}"))

    return port_tag_tuples



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import enum
from typing import List, Optional, TypeVar

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool._core.class_creation.attrs import attrs_optional_ndarray_isclose
from acconeer.exptool.a121.algo import (
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    ProcessorBase,
    double_buffering_frame_filter,
)


T = TypeVar("T", float, npt.NDArray[np.float64])


class MeasurementType(AlgoParamEnum):
    CLOSE_RANGE = enum.auto()
    FAR_RANGE = enum.auto()
    CLOSE_AND_FAR_RANGE = enum.auto()


@attributes_doc
@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    measurement_type: MeasurementType = attrs.field(
        default=MeasurementType.CLOSE_RANGE,
        converter=MeasurementType,
    )
    """The measurement type.
    This decides which of the other parameters to take into consideration.
    ``CLOSE_RANGE`` will apply the close range parameters onto the sweep.
    ``FAR_RANGE`` will apply the far range parameters onto the sweep.
    ``CLOSE_AND_FAR_RANGE`` will apply the parameters for close range onto the first subsweep and
    the parameters for far range onto the second subsweep."""

    sensitivity_close: float = attrs.field(default=1.9)
    """Sensitivity for close range detection. High sensitivity equals low detection threshold,
    low sensitivity equals high detection threshold."""

    sensitivity_far: float = attrs.field(default=2.0)
    """Sensitivity for far range detection. High sensitivity equals low detection threshold,
    low sensitivity equals high detection threshold."""

    patience_close: int = attrs.field(default=2)
    """Number of frames in a row above threshold to count as a new close range detection,
    also number of frames in a row below threshold to count as end of detection."""

    patience_far: int = attrs.field(default=2)
    """Number of frames in a row above threshold to count as a new far range detection,
    also number of frames in a row below threshold to count as end of detection."""

    calibration_duration_s: float = attrs.field(default=0.6)
    """Calibration duration in seconds"""

    calibration_interval_s: float = attrs.field(default=20)
    """Interval between calibrations in seconds. When reached a new calibration is made.
    Should not be set lower than the longest estimated continuous detection event."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if (
            config.sensor_config.num_subsweeps != 2
            and self.measurement_type is MeasurementType.CLOSE_AND_FAR_RANGE
        ):
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "num_subsweeps",
                    'Number of subsweeps must be 2 for range "Close and far"',
                )
            )

        if config.sensor_config.num_subsweeps != 1 and (
            self.measurement_type is MeasurementType.CLOSE_RANGE
            or self.measurement_type is MeasurementType.FAR_RANGE
        ):
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "num_subsweeps",
                    'Number of subsweeps must be 1 for ranges "Close" and "Far"',
                )
            )

        if config.sensor_config.sweep_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweep_rate",
                    "Sweep rate must be set.",
                )
            )

        else:
            if config.sensor_config.sweeps_per_frame > (
                self.calibration_duration_s * config.sensor_config.sweep_rate
            ):
                calibration_limit = np.around(
                    config.sensor_config.sweeps_per_frame / config.sensor_config.sweep_rate, 2
                )
                validation_results.append(
                    a121.ValidationError(
                        self,
                        "calibration_duration_s",
                        (
                            f"Calibration duration must be at least {calibration_limit} s. "
                            "Following condition applies:\n"
                            "sweeps per frame > (calibration duration * sweep rate)"
                        ),
                    )
                )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorResult:
    close: Optional[RangeResult] = attrs.field(default=None)
    """Returns the ``RangeResult`` for close range. ``None`` if range is not activated."""

    far: Optional[RangeResult] = attrs.field(default=None)
    """Returns the ``RangeResult`` for far range. ``None`` if range is not activated."""


@attrs.frozen(kw_only=True)
class RangeResult:
    detection: bool = attrs.field()
    """Detection in current range. ``True`` if detection and ``False`` if no detection."""

    threshold: float = attrs.field()
    """Detection score threshold.
    Depends on the sensitivity parameter for the current range, the threshold is
    equal to 10 / *sensitivity*."""

    score: npt.NDArray[np.float64] = attrs.field(eq=attrs_optional_ndarray_isclose)
    """Detection score for each point and sweep in current range.
    The output has the shape of (sweeps per frame, number of points in current subsweep)."""


class Processor(ProcessorBase[ProcessorResult]):
    """Touchless Button processor

    :param sensor_config: Sensor configuration
    :param metadata: Metadata yielded by the sensor config
    :param processor_config: Processor configuration
    """

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
    ) -> None:
        self._sensor_config = sensor_config
        self._metadata = metadata
        self._processor_config = processor_config

        self._processor_config.validate(self._sensor_config)

        self._sweeps_per_frame = sensor_config.sweeps_per_frame

        if sensor_config.sweep_rate is None:
            msg = "sweep_rate must be set"
            raise ValueError(msg)

        frame_rate = sensor_config.sweep_rate / sensor_config.sweeps_per_frame
        self._cal_interval_frames = int(processor_config.calibration_interval_s * frame_rate)

        self._reset_background()

        self._detection_close = False
        self._detection_far = False

        self._sig_count = np.zeros((2,))
        self._nonsig_count = np.array(
            [processor_config.patience_close + 1, processor_config.patience_far + 1],
        )

    def process(self, result: a121.Result) -> ProcessorResult:
        if self._sensor_config.double_buffering:
            filter_output = double_buffering_frame_filter(result._frame)
            if filter_output is None:
                frame = result.frame
            else:
                frame = filter_output
        else:
            frame = result.frame

        y = np.zeros_like(frame, dtype=float)

        sensitivity = self._get_sensitivity()

        if self._frames_since_last_cal > self._cal_interval_frames:
            self._reset_background()
        elif not np.any(np.isnan(self._dynamic_background)):
            y = self._calc_variance(frame)

        threshold = self._get_threshold_from_sensitivity(sensitivity)
        significant = y > threshold
        detection_depth, counts = np.unique(np.where(significant)[1], return_counts=True)

        if self._processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            self._detection_close = self._process_single_range(
                frame=frame,
                detection=self._detection_close,
                patience=self._processor_config.patience_close,
                detection_depth=detection_depth,
                counts=counts,
                measurement_type=MeasurementType.CLOSE_RANGE,
            )
            if self._processor_config.sensitivity_close is None:
                threshold_close = None
            else:
                threshold_close = self._get_threshold_from_sensitivity(
                    self._processor_config.sensitivity_close
                )
            return ProcessorResult(
                close=RangeResult(
                    detection=self._detection_close,
                    threshold=threshold_close,
                    score=y,
                )
            )

        if self._processor_config.measurement_type == MeasurementType.FAR_RANGE:
            self._detection_far = self._process_single_range(
                frame=frame,
                detection=self._detection_far,
                patience=self._processor_config.patience_far,
                detection_depth=detection_depth,
                counts=counts,
                measurement_type=MeasurementType.FAR_RANGE,
            )
            if self._processor_config.sensitivity_far is None:
                threshold_far = None
            else:
                threshold_far = self._get_threshold_from_sensitivity(
                    self._processor_config.sensitivity_far
                )
            return ProcessorResult(
                far=RangeResult(
                    detection=self._detection_far,
                    threshold=threshold_far,
                    score=y,
                )
            )

        if self._processor_config.measurement_type == MeasurementType.CLOSE_AND_FAR_RANGE:
            [self._detection_close, self._detection_far] = self._process_multiple_ranges(
                frame=frame,
                detection_depth=detection_depth,
                counts=counts,
            )
            score_close, score_far = np.split(
                y, [self._sensor_config.subsweeps[0].num_points], axis=1
            )
            return ProcessorResult(
                close=RangeResult(
                    detection=self._detection_close,
                    threshold=self._get_threshold_from_sensitivity(
                        self._processor_config.sensitivity_close
                    ),
                    score=score_close,
                ),
                far=RangeResult(
                    detection=self._detection_far,
                    threshold=self._get_threshold_from_sensitivity(
                        self._processor_config.sensitivity_far
                    ),
                    score=score_far,
                ),
            )

        raise AssertionError

    def _reset_background(self) -> None:
        assert self._sensor_config.sweep_rate is not None
        self._dynamic_background = np.full(
            (
                int(
                    self._processor_config.calibration_duration_s * self._sensor_config.sweep_rate
                ),
                self._metadata.sweep_data_length,
            ),
            np.nan,
            dtype="complex",
        )
        self._dynamic_background_guard = np.full(
            (self._sweeps_per_frame, self._metadata.sweep_data_length),
            np.nan,
            dtype="complex",
        )
        self._frames_since_last_cal = 0

    def _calc_variance(self, frame: npt.NDArray[np.complex128]) -> npt.NDArray[np.float64]:
        xn = np.full((self._sweeps_per_frame, self._metadata.sweep_data_length), 0, dtype=float)
        y = np.full((self._sweeps_per_frame, self._metadata.sweep_data_length), 0, dtype=float)

        arg_norm = np.mean(self._dynamic_background, axis=0)
        arg_norm = np.conj(arg_norm) / np.abs(arg_norm)

        arg_norm_ref = self._dynamic_background * arg_norm
        ref_ampls = np.abs(arg_norm_ref)
        ampl_mean = np.mean(ref_ampls, axis=0)
        ampl_std = np.std(ref_ampls, axis=0)
        ref_phases = np.angle(arg_norm_ref)
        phase_mean = np.mean(ref_phases, axis=0)
        phase_std = np.std(ref_phases, axis=0)

        xn = frame * arg_norm

        y = np.hypot(
            (np.abs(xn) - ampl_mean) / ampl_std,
            (np.angle(xn) - phase_mean) / phase_std,
        )

        return y

    def _process_single_range(
        self,
        frame: npt.NDArray[np.complex128],
        detection: bool,
        patience: int,
        detection_depth: npt.NDArray[np.int_],
        counts: npt.NDArray[np.int_],
        measurement_type: MeasurementType,
    ) -> bool:
        if measurement_type == MeasurementType.CLOSE_RANGE:
            index = 0
        elif measurement_type == MeasurementType.FAR_RANGE:
            index = 1
        else:
            raise AssertionError

        # Checks if there are at least two significant values at the same depth
        if detection_depth[counts > 1].size > 0:
            self._sig_count[index] += 1
            self._nonsig_count[index] = 0
            self._dynamic_background_guard = np.full(
                (
                    self._sensor_config.sweeps_per_frame,
                    self._metadata.sweep_data_length,
                ),
                np.nan,
                dtype="complex",
            )
            self._frames_since_last_cal += 1
        else:
            self._sig_count[index] = 0
            self._nonsig_count[index] += 1

            if not np.isnan(self._dynamic_background_guard).any():
                self._update_background()

            self._dynamic_background_guard = frame

        detection = self._get_detection(
            detection,
            self._sig_count[index],
            self._nonsig_count[index],
            patience,
        )

        return detection

    def _process_multiple_ranges(
        self,
        frame: npt.NDArray[np.complex128],
        detection_depth: npt.NDArray[np.int_],
        counts: npt.NDArray[np.int_],
    ) -> List[bool]:
        if detection_depth[counts > 1].size > 0:
            if (
                np.where(
                    detection_depth[counts > 1] < self._sensor_config.subsweeps[0].num_points,
                    True,
                    False,
                )
            ).any():
                self._sig_count[0] += 1
                self._nonsig_count[0] = 0
            else:
                self._sig_count[0] = 0
                self._nonsig_count[0] += 1
            if (
                np.where(
                    detection_depth[counts > 1] >= self._sensor_config.subsweeps[0].num_points,
                    True,
                    False,
                )
            ).any():
                self._sig_count[1] += 1
                self._nonsig_count[1] = 0
            else:
                self._sig_count[1] = 0
                self._nonsig_count[1] += 1

            self._dynamic_background_guard = np.full(
                (
                    self._sensor_config.sweeps_per_frame,
                    self._metadata.sweep_data_length,
                ),
                np.nan,
                dtype="complex",
            )
            self._frames_since_last_cal += 1

        else:
            self._sig_count[0] = 0
            self._nonsig_count[0] += 1
            self._sig_count[1] = 0
            self._nonsig_count[1] += 1

            if not np.isnan(self._dynamic_background_guard).any():
                self._update_background()

            self._dynamic_background_guard = frame

        detection_close = self._get_detection(
            self._detection_close,
            self._sig_count[0],
            self._nonsig_count[0],
            self._processor_config.patience_close,
        )

        detection_far = self._get_detection(
            self._detection_far,
            self._sig_count[1],
            self._nonsig_count[1],
            self._processor_config.patience_far,
        )

        return [detection_close, detection_far]

    def _get_sensitivity(self) -> npt.NDArray[np.float64]:
        if self._processor_config.measurement_type == MeasurementType.FAR_RANGE:
            return np.repeat(
                self._processor_config.sensitivity_far,
                self._sensor_config.num_points,
            )
        elif self._processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            return np.repeat(
                self._processor_config.sensitivity_close,
                self._sensor_config.num_points,
            )
        elif self._processor_config.measurement_type == MeasurementType.CLOSE_AND_FAR_RANGE:
            return np.repeat(
                [
                    self._processor_config.sensitivity_close,
                    self._processor_config.sensitivity_far,
                ],
                [
                    self._sensor_config.subsweeps[0].num_points,
                    self._sensor_config.subsweeps[1].num_points,
                ],
            )

        raise AssertionError

    def _update_background(self) -> None:
        self._dynamic_background = np.roll(
            self._dynamic_background, -self._sweeps_per_frame, axis=0
        )
        self._dynamic_background[-self._sweeps_per_frame :, :] = self._dynamic_background_guard
        self._frames_since_last_cal = 0

    @staticmethod
    def _get_threshold_from_sensitivity(sensitivity: T) -> T:
        threshold = 1 / sensitivity * 10
        return threshold

    @staticmethod
    def _get_detection(
        curr_detection: object, sig_count: int, nonsig_count: int, patience: int
    ) -> bool:
        new_detection = not curr_detection and sig_count >= patience
        keep_detection = curr_detection and nonsig_count <= patience
        detection = new_detection or keep_detection
        return bool(detection)


def _load_algo_data(algo_group: h5py.Group) -> ProcessorConfig:
    processor_config = ProcessorConfig.from_json(algo_group["processor_config"][()])
    return processor_config



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved
from __future__ import annotations

import socket
from time import monotonic
from typing import Optional

from .buffered_link import BufferedLink, LinkError


class SocketLink(BufferedLink):
    _CHUNK_SIZE = 4096
    _PORT = 6110

    def __init__(self, host: Optional[str] = None, port: Optional[int] = None) -> None:
        super().__init__()
        self._host = host
        self._sock: Optional[socket.socket] = None
        self._buf: bytearray = bytearray()
        self._port: int = self._PORT if (port is None) else port

    def _update_timeout(self) -> None:
        if self._sock is not None:
            self._sock.settimeout(self._timeout)

    def connect(self) -> None:
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._update_timeout()

        try:
            self._sock.connect((self._host, self._port))
        except OSError as e:
            self._sock.close()
            self._sock = None
            msg = "failed to connect"
            raise LinkError(msg) from e

        self._buf = bytearray()

    def recv(self, num_bytes: int) -> bytes:
        assert self._sock is not None
        while len(self._buf) < num_bytes:
            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        data = self._buf[:num_bytes]
        self._buf = self._buf[num_bytes:]
        return data

    def recv_until(self, bs: bytes) -> bytes:
        assert self._sock is not None
        t0 = monotonic()
        while True:
            try:
                i = self._buf.index(bs)
            except ValueError:
                pass
            else:
                break

            if monotonic() - t0 > self._timeout:
                msg = "recv timeout"
                raise LinkError(msg)

            try:
                r = self._sock.recv(self._CHUNK_SIZE)
            except OSError as e:
                raise LinkError from e
            self._buf.extend(r)

        i += 1
        data = self._buf[:i]
        self._buf = self._buf[i:]

        return data

    def send(self, data: bytes) -> None:
        assert self._sock is not None
        self._sock.sendall(data)

    def disconnect(self) -> None:
        if self._sock is not None:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
            self._sock = None
        self._buf = bytearray()



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from .metadata_view import ExtendedMetadataView, MetadataView, SmartMetadataView
from .perf_calc_view import ExtendedPerfCalcView, PerfCalcView, SmartPerfCalcView
from .range_help_view import RangeHelpView
from .sensor_config_editor import SensorConfigEditor
from .session_config_editor import SessionConfigEditor
from .subsweep_config_editor import SubsweepConfigEditor



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import contextlib
import logging
import sys
import traceback
from typing import Iterator, Optional

import pyperclip

from PySide6 import QtGui
from PySide6.QtWidgets import QFrame, QHBoxLayout, QMessageBox, QPushButton, QWidget

from acconeer.exptool.app.new._exceptions import HandledException


_LOG = logging.getLogger(__name__)


def get_installed_packages_via_pip() -> Optional[list[str]]:
    """Tries to extract information about installed packages"""
    try:
        try:
            from pip._internal.operations.freeze import freeze
        except ImportError:
            from pip.operations.freeze import freeze  # type: ignore

        return list(freeze())
    except Exception as e:
        _LOG.exception(e)
        return None


class ExceptionWidget(QMessageBox):
    def __init__(
        self,
        parent: Optional[QWidget],
        *,
        exc: Exception,
        traceback_str: Optional[str] = None,
        title: str = "Error",
    ) -> None:
        super().__init__(parent)

        self.setIcon(QMessageBox.Icon.Warning)
        self.setStandardButtons(QMessageBox.StandardButton.Ok)

        self.setWindowTitle(title)
        self.setText(str(exc))

        try:
            raise exc
        except HandledException:
            pass
        except Exception:
            self.setInformativeText("<b>Unhandled error - please file a bug</b>")

        # Monospaced detailed text.
        self.setStyleSheet("QTextEdit { font-family: monospace; font-size: 9pt; }")

        detailed_text = (
            "Checklist before reporting a bug\n"
            + "================================\n"
            + "- [ ] I have read the 'Common Issues' without finding a solution\n"
            + "      (See https://docs.acconeer.com/en/latest/exploration_tool/faq.html)\n"
            + "- [ ] This issue occured in the latest release of Exploration Tool\n"
            + "      (See https://pypi.org/project/acconeer-exptool/)\n"
            + "- [ ] I have double-checked that the section 'Installed Packages' contains\n"
            + "      rows on the format '<package name>==<version>'\n"
            + "\n\n"
        )

        detailed_text += f"Python version\n==============\n{sys.version}\n\n\n"

        installed_packages = get_installed_packages_via_pip()
        detailed_text += "Installed Packages\n" + "==================\n"
        if installed_packages is None:
            detailed_text += (
                "Please supply the output of 'pip freeze'\n"
                + "(Could not extract information about installed packages)."
            )
        else:
            detailed_text += "\n".join(installed_packages)
        detailed_text += "\n\n\n"

        if traceback_str:
            detailed_text += "Traceback\n" + "=========\n" + traceback_str

        if detailed_text:
            self.setDetailedText(detailed_text)
            copy_button = QPushButton(self)
            copy_button.setText("Copy details")
            self.addButton(copy_button, QMessageBox.ButtonRole.ActionRole)
            copy_button.clicked.disconnect()
            copy_button.clicked.connect(self._on_copy_clicked)

    @classmethod
    @contextlib.contextmanager
    def context(cls) -> Iterator[None]:
        try:
            yield
        except Exception as e:
            _LOG.debug("Exception raised in MainThread:", exc_info=True)
            cls(parent=None, exc=e, traceback_str=traceback.format_exc()).exec()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self.setMinimumWidth(600)

    def _on_copy_clicked(self) -> None:
        detailed_text = self.detailedText()
        if detailed_text:
            pyperclip.copy(detailed_text)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.close()


class VerticalSeparator(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(5, 5, 5, 5)

        frame = QFrame(self)
        frame.setFrameShape(QFrame.Shape.VLine)
        layout.addWidget(frame)

        self.setLayout(layout)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from .load_dialog import LoadDialogWithJsonEditor
from .preview import PresentationType, PresentationWindow, PresenterFunc
from .save_dialog import SaveDialogWithPreview
from .set_config_presenter import set_config_presenter



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved
r"""
This module contains the power modelling building blocks.

The main idea is that an energy region has a duration (seconds)
and an average current (ampere). This means that all energy regions
can be plotted on a 2d plot.

    A
    ^
    |
    |   +---------------+  \
    |   |               |   \_Average current
    |   | energy region |   /
    |   |               |  /
    +-------------------------> t

        |-- duration  --|

A rectangle like this is called a simple region.

In order to get more advanced regions, the simple regions
can be composed:

    A
    ^
    |         +-+
    |         | |
    |         |s|
    |         |r|-----+ ____Average current
    |   +-----|2| sr3 |
    |   | sr1 | |     |
    +-------------------------> t

        |seq. duration|

These are called composte regions.

P.S. Composite regions can be composed from other composite regions.
"""

from __future__ import annotations

import itertools
import operator
import typing as t
from enum import Enum, auto

import attrs
import typing_extensions as te


_T = t.TypeVar("_T")


class EnergyRegion(te.Protocol):
    """
    Interface of energy regions
    """

    class Tag(Enum):
        MEASURE = auto()
        CALIBRATION = auto()
        IDLE = auto()
        OVERHEAD = auto()
        PROCESSING = auto()
        COMMUNICATION = auto()

    @property
    def average_current(self) -> float: ...

    @property
    def duration(self) -> float: ...

    @property
    def charge(self) -> float:
        return self.average_current * self.duration

    @property
    def description(self) -> str: ...

    def truncate(self, new_duration: float) -> EnergyRegion:
        """Returns a (time-)truncated copy of this region"""
        ...

    def join(self, regions: t.Iterable[EnergyRegion], description: str = "") -> EnergyRegion:
        """
        Helper function. See `_join`
        """
        regions = tuple(regions)

        if len(regions) == 0:
            msg = "Cannot join an empty sequence."
            raise ValueError(msg)
        elif len(regions) == 1:
            (region,) = regions
            return region
        else:
            return CompositeRegion(tuple(_join(self, regions)), description)


@attrs.frozen
class SimpleRegion(EnergyRegion):
    """A simple region"""

    current: float
    duration: float = attrs.field(validator=attrs.validators.ge(0.0))
    tag: t.Optional[EnergyRegion.Tag] = None
    description: str = ""

    @property
    def average_current(self) -> float:
        return self.current

    def truncate(self, new_duration: float) -> SimpleRegion:
        return attrs.evolve(
            self,
            duration=new_duration,
            description=f"{self.description} (truncated)",
        )


@attrs.frozen
class CompositeRegion(EnergyRegion):
    """
    Composes energy regions in a tree-structure.
    """

    regions: tuple[EnergyRegion, ...]
    description: str = ""

    @property
    def average_current(self) -> float:
        return sum(r.average_current * r.duration / self.duration for r in self.regions)

    @property
    def duration(self) -> float:
        return duration(*self.regions)

    def truncate(self, new_duration: float) -> CompositeRegion:
        durations = (r.duration for r in self.regions)
        starts = itertools.accumulate(durations, operator.add)
        index_to_truncate = sum(start < new_duration for start in starts)

        whole_regions = self.regions[:index_to_truncate]
        whole_regions_duration = sum(r.duration for r in whole_regions)
        if index_to_truncate < len(self.regions):
            truncated_region = self.regions[index_to_truncate].truncate(
                new_duration - whole_regions_duration
            )
            return CompositeRegion(
                whole_regions + (truncated_region,), description=self.description
            )
        else:
            return CompositeRegion(whole_regions, description=self.description)

    def flat_iter(self) -> t.Iterator[SimpleRegion]:
        """
        Returns an in-order iterator of all the SimpleRegions (leaf nodes).
        """
        for r in self.regions:
            if isinstance(r, SimpleRegion):
                yield r
            elif isinstance(r, CompositeRegion):
                yield from r.flat_iter()


def duration(*regions: EnergyRegion) -> float:
    """
    Returns the total duration of a variable amount of regions
    """
    return sum(r.duration for r in regions)


def _join(joiner: _T, iterable: t.Iterable[_T]) -> t.Iterator[_T]:
    """
    Similar to str.join (except the last concat step) but for any type.
    """
    iterator = iter(iterable)
    yield next(iterator)

    for e in iterator:
        yield joiner
        yield e



# Copyright (c) Acconeer AB, 2024
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Callable, Optional, Type

import numpy as np

from PySide6 import QtCore

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
    ProcessorPluginPreset,
    ProcessorViewPluginBase,
    SetupMessage,
)
from acconeer.exptool.a121.algo._utils import APPROX_BASE_STEP_LENGTH_M, get_distances_m
from acconeer.exptool.a121.algo.waste_level import (
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_processor_config,
    get_sensor_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PgPlotPlugin,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    pidgets,
)


log = logging.getLogger(__name__)


class PluginPresetId(Enum):
    PLASTIC_WASTE_BIN = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):
    PLUGIN_PRESETS = {
        PluginPresetId.PLASTIC_WASTE_BIN.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_sensor_config()),
            processor_config=get_processor_config(),
        ),
    }

    @classmethod
    def get_processor(cls, state: ProcessorBackendPluginSharedState[ProcessorConfig]) -> Processor:
        if state.metadata is None:
            msg = "metadata is None"
            raise RuntimeError(msg)

        if isinstance(state.metadata, list):
            msg = "metadata is unexpectedly extended"
            raise RuntimeError(msg)

        return Processor(
            sensor_config=state.session_config.sensor_config,
            processor_config=state.processor_config,
            metadata=state.metadata,
        )

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "bin_start_m": pidgets.FloatPidgetFactory(
                name_label_text="Bin start:",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "bin_start_m"),
                suffix=" m",
                decimals=3,
                limits=(0.03, 20),
            ),
            "bin_end_m": pidgets.FloatPidgetFactory(
                name_label_text="Bin end:",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "bin_end_m"),
                suffix=" m",
                decimals=3,
                limits=(0.05, 20),
            ),
            "threshold": pidgets.FloatSliderPidgetFactory(
                name_label_text="Threshold:",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "threshold"),
                decimals=1,
                limits=(0.0, 4),
            ),
            "distance_sequence_n": pidgets.IntPidgetFactory(
                name_label_text="Number of distances in sequence:",
                name_label_tooltip=get_attribute_docstring(ProcessorConfig, "distance_sequence_n"),
                limits=(1, 10),
            ),
            "median_filter_length": pidgets.IntPidgetFactory(
                name_label_text="Median filter length:",
                name_label_tooltip=get_attribute_docstring(
                    ProcessorConfig, "median_filter_length"
                ),
                limits=(1, 10),
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def supports_multiple_subsweeps(self) -> bool:
        return True


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[ProcessorResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            if isinstance(message.metadata, list):
                msg = "Metadata is unexpectedly extended"
                raise RuntimeError(msg)

            self.setup(
                metadata=message.metadata,
                session_config=message.session_config,
                processor_config=message.processor_config,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(processor_result=self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        metadata: a121.Metadata,
        session_config: a121.SessionConfig,
        processor_config: ProcessorConfig,
    ) -> None:
        self.plot_layout.clear()

        # Phase standard deviation plot
        self.phase_std_plot = self._create_phase_std_plot(self.plot_layout)
        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        self.phase_std_curve = self.phase_std_plot.plot(pen=pen)
        self.phase_std_dots_above = pg.ScatterPlotItem(symbol="o", size=10, brush=brush, pen="k")
        self.phase_std_plot.addItem(self.phase_std_dots_above)
        brush = et.utils.pg_brush_cycler(1)
        self.phase_std_dots_below = pg.ScatterPlotItem(symbol="o", size=10, brush=brush, pen="k")
        self.phase_std_plot.addItem(self.phase_std_dots_below)
        brush = et.utils.pg_brush_cycler(2)
        self.phase_std_dots_detection = pg.ScatterPlotItem(
            symbol="o", size=10, brush=brush, pen="k"
        )
        self.phase_std_plot.addItem(self.phase_std_dots_detection)

        dashed_pen = pg.mkPen("k", width=2.5, style=QtCore.Qt.PenStyle.DashLine)
        threshold_line = pg.InfiniteLine(pos=processor_config.threshold, angle=0, pen=dashed_pen)
        self.phase_std_plot.addItem(threshold_line)

        vertical_line_start = pg.InfiniteLine(
            pos=processor_config.bin_start_m,
            pen=et.utils.pg_pen_cycler(7),
            label="Bin top",
            labelOpts={
                "position": 0.6,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.phase_std_plot.addItem(vertical_line_start)

        vertical_line_end = pg.InfiniteLine(
            pos=processor_config.bin_end_m,
            pen=et.utils.pg_pen_cycler(7),
            label="Bin bottom",
            labelOpts={
                "position": 0.6,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.phase_std_plot.addItem(vertical_line_end)

        self.level_line = pg.InfiniteLine(
            pen=et.utils.pg_pen_cycler(2),
            label="Fill level",
            labelOpts={
                "position": 0.8,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.phase_std_plot.addItem(self.level_line)
        self.level_line.hide()

        self.distances_m = get_distances_m(session_config.sensor_config, metadata)
        self.threshold = processor_config.threshold
        self.sequence_ones = np.ones(processor_config.distance_sequence_n)

        # Level history plot
        self.level_history_plot = self._create_history_plot(self.plot_layout)

        if session_config.sensor_config.frame_rate is not None:
            history_length_s = 5
            history_length_n = int(
                round(history_length_s * session_config.sensor_config.frame_rate)
            )
            self.hist_xs = np.linspace(-history_length_s, 0, history_length_n)
        elif session_config.update_rate is not None:
            history_length_s = 5
            history_length_n = int(round(history_length_s * session_config.update_rate))
            self.hist_xs = np.linspace(-history_length_s, 0, history_length_n)
        else:
            history_length_n = 100
            self.hist_xs = np.linspace(-history_length_n, 0, history_length_n)
            self.level_history_plot.setLabel("bottom", "Frame")

        self.level_history = np.full(history_length_n, np.nan)
        self.level_history_plot.setYRange(
            0,
            processor_config.bin_end_m
            - np.minimum(
                processor_config.bin_start_m,
                session_config.sensor_config.subsweeps[0].start_point * APPROX_BASE_STEP_LENGTH_M,
            ),
        )

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.level_history_curve = self.level_history_plot.plot(**feat_kw, connect="finite")

        top_bin_horizontal_line = pg.InfiniteLine(
            pos=processor_config.bin_end_m - processor_config.bin_start_m,
            pen=et.utils.pg_pen_cycler(7),
            angle=0,
            label="Bin top",
            labelOpts={
                "position": 0.5,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.level_history_plot.addItem(top_bin_horizontal_line)

        bottom_bin_horizontal_line = pg.InfiniteLine(
            pos=0,
            pen=et.utils.pg_pen_cycler(7),
            angle=0,
            label="Bin bottom",
            labelOpts={
                "position": 0.5,
                "color": (0, 100, 0),
                "fill": (200, 200, 200, 50),
                "movable": True,
            },
        )
        self.level_history_plot.addItem(bottom_bin_horizontal_line)

        # Level plot
        self.num_rects = 16
        self.rect_plot = pg.PlotItem()
        self.rect_plot.setAspectLocked()
        self.rect_plot.hideAxis("left")
        self.rect_plot.hideAxis("bottom")
        self.rects = []

        pen = pg.mkPen(None)
        rect_width = self.num_rects / 2.0
        for r in np.arange(self.num_rects) + 1:
            rect = pg.QtWidgets.QGraphicsRectItem(0, r, rect_width, 1)
            rect.setPen(pen)
            self.rect_plot.addItem(rect)
            self.rects.append(rect)

        self.level_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:12pt;">'
            "{}</span></div>"
        )

        self.level_text_item = pg.TextItem(
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        no_detection_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:12pt;">'
            "{}</span></div>".format("No detection")
        )

        self.no_detection_text_item = pg.TextItem(
            html=no_detection_html,
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )

        self.rect_plot.addItem(self.level_text_item)
        self.rect_plot.addItem(self.no_detection_text_item)
        self.level_text_item.setPos(self.num_rects / 4.0, self.num_rects + 4.0)
        self.level_text_item.hide()
        self.no_detection_text_item.setPos(self.num_rects / 4.0, self.num_rects + 4.0)
        self.no_detection_text_item.show()

        self.plot_layout.addItem(self.rect_plot, row=0, col=0)

    def draw_plot_job(self, processor_result: ProcessorResult) -> None:
        # Phase standard deviation plot
        self.phase_std_curve.setData(self.distances_m, processor_result.extra_result.phase_std)
        if processor_result.extra_result.distance_m is not None:
            self.level_line.setPos(processor_result.extra_result.distance_m)
            self.level_line.show()
        else:
            self.level_line.hide()

        detection_array = processor_result.extra_result.phase_std < self.threshold
        if np.all(detection_array):
            self.phase_std_dots_below.setData(
                self.distances_m, processor_result.extra_result.phase_std
            )
            self.phase_std_dots_above.hide()
            self.phase_std_dots_detection.hide()
        elif np.any(detection_array):
            above_idxs = np.argwhere(~detection_array)
            above_idxs = above_idxs.reshape(above_idxs.shape[0])
            below_idxs = np.argwhere(detection_array)
            below_idxs = below_idxs.reshape(below_idxs.shape[0])
            consecutive_true_indices = np.where(
                np.convolve(detection_array, self.sequence_ones, mode="valid")
                == self.sequence_ones.shape[0]
            )[0]
            detection_idxs = []
            for i in consecutive_true_indices:
                for j in np.arange(i, i + self.sequence_ones.shape[0]):
                    if j not in detection_idxs and j < detection_array.shape[0]:
                        detection_idxs.append(j)

            remove = np.isin(below_idxs, detection_idxs)
            below_idxs = below_idxs[~remove]

            self.phase_std_dots_above.setData(
                self.distances_m[above_idxs], processor_result.extra_result.phase_std[above_idxs]
            )
            self.phase_std_dots_below.setData(
                self.distances_m[below_idxs], processor_result.extra_result.phase_std[below_idxs]
            )
            if len(detection_idxs) > 0:
                self.phase_std_dots_detection.setData(
                    self.distances_m[detection_idxs],
                    processor_result.extra_result.phase_std[detection_idxs],
                )
                self.phase_std_dots_detection.show()
            else:
                self.phase_std_dots_detection.hide()

            self.phase_std_dots_above.show()
            self.phase_std_dots_below.show()
        else:
            self.phase_std_dots_above.setData(
                self.distances_m, processor_result.extra_result.phase_std
            )
            self.phase_std_dots_below.hide()
            self.phase_std_dots_detection.hide()

        # History plot

        self.level_history = np.roll(self.level_history, -1)
        if processor_result.level_m is not None:
            self.level_history[-1] = processor_result.level_m
        else:
            self.level_history[-1] = np.nan

        if np.all(np.isnan(self.level_history)):
            self.level_history_curve.hide()
        else:
            self.level_history_curve.setData(self.hist_xs, self.level_history)
            self.level_history_curve.show()

        # Level plot

        # Show the percentage level plot if the plot width is greater than 600 pixels,
        # otherwise display the level as text.
        if self.plot_layout.width() < 600:
            if processor_result.level_percent is None:
                self.level_text_item.hide()
                self.no_detection_text_item.show()
            elif processor_result.level_percent > 100:  # Overflow
                level_text = "Overflow"
                level_html = self.level_html_format.format(level_text)
                self.level_text_item.setHtml(level_html)
                self.level_text_item.show()
                self.no_detection_text_item.hide()
            elif processor_result.level_percent > 0:  # In bin detection
                assert processor_result.level_m is not None
                assert processor_result.level_percent is not None
                level_text = "Level: {:.2f} m, {:.0f} %".format(
                    processor_result.level_m,
                    processor_result.level_percent,
                )
                level_html = self.level_html_format.format(level_text)
                self.level_text_item.setHtml(level_html)
                self.level_text_item.show()
                self.no_detection_text_item.hide()
            else:  # No detection
                self.level_text_item.hide()
                self.no_detection_text_item.show()

            for rect in self.rects:
                rect.setVisible(False)
        else:
            if processor_result.level_percent is None:  # No detection
                level_text = "No detection"
                for rect in self.rects:
                    rect.setBrush(et.utils.pg_brush_cycler(7))
                self.level_text_item.hide()
                self.no_detection_text_item.show()
            elif processor_result.level_percent > 100:  # Overflow
                for rect in self.rects:
                    rect.setBrush(et.utils.pg_brush_cycler(0))

                level_text = "Overflow"
                level_html = self.level_html_format.format(level_text)
                self.level_text_item.setHtml(level_html)
                self.level_text_item.show()
                self.no_detection_text_item.hide()
            else:  # In bin detection
                self.bar_loc = int(
                    np.around(processor_result.level_percent / 100 * self.num_rects)
                )
                for rect in self.rects[: self.bar_loc]:
                    rect.setBrush(et.utils.pg_brush_cycler(0))

                for rect in self.rects[self.bar_loc :]:
                    rect.setBrush(et.utils.pg_brush_cycler(7))

                assert processor_result.level_m is not None
                assert processor_result.level_percent is not None
                level_text = "Level: {:.2f} m, {:.0f} %".format(
                    processor_result.level_m,
                    processor_result.level_percent,
                )
                level_html = self.level_html_format.format(level_text)
                self.level_text_item.setHtml(level_html)
                self.level_text_item.show()
                self.no_detection_text_item.hide()

            for rect in self.rects:
                rect.setVisible(True)

    @staticmethod
    def _create_phase_std_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        phase_std_plot = parent.addPlot(row=1, col=0, colspan=3)
        phase_std_plot.setTitle("Phase standard deviation")
        phase_std_plot.setLabel(axis="bottom", text="Distance [m]")
        phase_std_plot.setLabel("left", "Phase std")
        phase_std_plot.setMenuEnabled(False)
        phase_std_plot.setMouseEnabled(x=False, y=False)
        phase_std_plot.hideButtons()
        phase_std_plot.showGrid(x=True, y=True, alpha=0.5)
        phase_std_plot.setYRange(0, 4)

        return phase_std_plot

    @staticmethod
    def _create_history_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        history_plot = parent.addPlot(row=0, col=1, colspan=2)
        history_plot.setTitle("Level history")
        history_plot.setMenuEnabled(False)
        history_plot.showGrid(x=True, y=True)
        history_plot.setLabel("left", "Estimated level (m)")
        history_plot.setLabel("bottom", "Time (s)")

        return history_plot


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


WASTE_LEVEL_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="waste_level",
    title="Waste level",
    docs_link="https://docs.acconeer.com/en/latest/example_apps/a121/waste_level.html",
    description="Detect waste level in a bin.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(
            name="Plastic waste bin",
            description="Settings for measuring waste level in a plastic waste bin (240 liter)",
            preset_id=PluginPresetId.PLASTIC_WASTE_BIN,
        ),
    ],
    default_preset_id=PluginPresetId.PLASTIC_WASTE_BIN,
)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

import base64
import hashlib
from enum import Enum
from struct import pack, unpack_from
from typing import Any, Callable, Dict, List, Optional, Tuple

import cbor2
import serial

from acconeer.exptool._core.communication.comm_devices import SerialDevice
from acconeer.exptool.flash._device_flasher_base import DeviceFlasherBase


MCUMGR_DEFAULT_BAUDRATE = 115200
MCUMGR_UART_TIMEOUT = 2

DFU_BOOT_DESCRIPTION = {
    "XM126": (
        "<p>"
        "<ol>"
        "<li>Press and hold the <b>DFU</b> button on the board</li>"
        "<li>Press the <b>RESET</b> button (still holding the DFU button)</li>"
        "<li>Release the <b>RESET</b> button</li>"
        "<li>Release the <b>DFU</b> button</li>"
        "</ol>"
        "</p>"
    ),
}


class McuMgrFlashException(Exception):
    pass


class McuMgrUartFlasher(DeviceFlasherBase):
    def flash(
        self,
        serial_device: SerialDevice,
        device_name: str,
        image_path: str,
        progress_callback: Optional[Callable[[int, bool], None]] = None,
    ) -> None:
        flasher = McuMgrFlashProtocol(serial_device.port)

        with open(image_path, "rb") as image_file:
            image_data = image_file.read()
            flasher.write_image(image_data, progress_callback=progress_callback)
            flasher.reset()

        flasher.close()

    @staticmethod
    def get_boot_description(device_name: str) -> Optional[str]:
        description = (
            "Please visit https://developer.acconeer.com to find the latest flash "
            f"instructions for {device_name}."
        )

        if device_name in DFU_BOOT_DESCRIPTION:
            description = (
                f"<p>To flash the {device_name} it needs to be put in bootloader mode:</p>"
                f"{DFU_BOOT_DESCRIPTION[device_name]}"
            )

        return description


class NmpOp(Enum):
    READ = 0
    READ_RSP = 1
    WRITE = 2
    WRITE_RSP = 3


class NmpGroup(Enum):
    OS = 0
    IMAGE = 1
    STAT = 2
    CONFIG = 3
    LOG = 4
    CRASH = 5
    SPLIT = 6
    RUN = 7
    FS = 8
    PERUSER = 64


class NmpIdOs(Enum):
    ECHO = 0
    CONS_ECHO_CTRL = 1
    TASKSTAT = 2
    MPSTAT = 3
    DATETIME_STR = 4
    RESET = 5


class NmpIdImage(Enum):
    STATE = 0
    UPLOAD = 1
    FILE = 2
    CORELIST = 3
    CORELOAD = 4
    ERASE = 5


class NmPHeader:
    fmt = "!BBHHBB"
    size = 8

    def __init__(
        self,
        op: NmpOp,
        group: NmpGroup,
        id: Enum,
        length: int = 0,
        seq_id: int = 0,
        flags: int = 0,
    ):
        self.op = op
        self.group = group
        self.id = id
        self.flags = flags
        self.length = length
        self.seq_id = seq_id

    @staticmethod
    def decode(b: bytes) -> NmPHeader:
        t = unpack_from(NmPHeader.fmt, b)

        return NmPHeader(t[0], t[3], t[5], length=t[2], seq_id=t[4], flags=t[1])

    def encode(self) -> bytes:
        return pack(
            NmPHeader.fmt,
            self.op.value,
            self.flags,
            self.length,
            self.group.value,
            self.seq_id,
            self.id.value,
        )

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}(op:{self.op} group:{self.group} id:{self.id} "
            f"len:{self.length} seq:{self.seq_id} flags:{self.flags})"
        )


class ResetRequestHeader(NmPHeader):
    def __init__(self, length: int = 0, seq_id: int = 0, flags: int = 0):
        super().__init__(NmpOp.WRITE, NmpGroup.OS, NmpIdOs.RESET, length, seq_id, flags)


class ImageRequestHeader(NmPHeader):
    def __init__(self, length: int = 0, seq_id: int = 0, flags: int = 0):
        super().__init__(NmpOp.WRITE, NmpGroup.IMAGE, NmpIdImage.UPLOAD, length, seq_id, flags)


class ResetRequest:
    def __init__(self, seq_id: int = 0):
        self.seq_id = seq_id

    def encode(self) -> bytes:
        req_header = ResetRequestHeader(length=1, seq_id=self.seq_id)

        data = req_header.encode()

        # calculate CRC16 of it and append to the request
        crc = McuMgrFlashProtocol.calc_crc16_xmodem(0, data).to_bytes(2, byteorder="big")
        data += crc

        # prepend chunk length
        chunk_len = len(data).to_bytes(2, byteorder="big")
        data = chunk_len + data

        # convert to base64
        enc_data = base64.b64encode(data)

        # Prepend start designator and append newline to base64 encoded data
        return b"\x06\x09" + enc_data + b"\n"


class ResetResponse:
    @staticmethod
    def decode(data: bytes, seq_id: int) -> None:
        # decode header and check seq
        header = ResetRequestHeader.decode(data)

        if seq_id != header.seq_id:
            msg = f"Seq number mismatch: {header}"
            raise McuMgrFlashException(msg)

        if len(data) > ResetRequestHeader.size:
            dec_msg: Dict[str, Any] = cbor2.loads(data[ResetRequestHeader.size :])
        else:
            msg = f"Complete header w/o payload: {header}"
            raise McuMgrFlashException(msg)

        if "rc" not in dec_msg:
            msg = "Missing key 'rc' in response"
            raise McuMgrFlashException(msg)

        if dec_msg["rc"] != 0:
            msg = f"Error code: '{dec_msg['rc']}' response"
            raise McuMgrFlashException(msg)


class ImageUploadRequest:
    def __init__(
        self,
        image_num: int,
        offset: int,
        data_len: Optional[int],
        data: bytes,
        data_sha: Optional[bytes],
        seq_id: int,
    ):
        self.offset = offset
        self.data_len = data_len
        self.data = data
        self.seq_id = seq_id

        if data_sha is None and data_len is None:
            self.payload_dict: Dict[str, Any] = {
                "off": self.offset,
                "data": self.data,
            }
        else:
            self.payload_dict = {
                "off": self.offset,
                "data": self.data,
                "sha": data_sha,
                "len": self.data_len,
            }

    def encode(self, line_length: int = 128) -> Tuple[List[bytes], int]:
        """
        encodes self.payload_dict and header
        returns both as bytes
        """
        ret_data = []

        # convert to bytes with CBOR
        payload_bytes = cbor2.dumps(self.payload_dict)

        req_header = ImageRequestHeader(length=len(payload_bytes), seq_id=self.seq_id)

        data = req_header.encode() + payload_bytes

        # calculate CRC16 of it and append to the request
        crc = McuMgrFlashProtocol.calc_crc16_xmodem(0, data).to_bytes(2, byteorder="big")
        data += crc

        # prepend chunk length
        chunk_len = len(data).to_bytes(2, byteorder="big")
        data = chunk_len + data

        # convert to base64
        enc_data = base64.b64encode(data)

        written = 0
        total_len = len(enc_data)

        # transfer in blocks of max line_length bytes per line
        while written < total_len:
            # start designator
            if written == 0:
                out_data = b"\x06\x09"
            else:
                out_data = b"\x04\x14"

            write_len = min(line_length - 4, total_len - written)
            out_data = out_data + enc_data[written : written + write_len] + b"\n"

            written = written + write_len

            ret_data.append(out_data)

        return ret_data, total_len

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(dict:{self.payload_dict}"


class ImageUploadResponse:
    @staticmethod
    def decode(data: bytes, seq_id: int) -> int:
        # decode header and check seq
        header = ImageRequestHeader.decode(data)

        if seq_id != header.seq_id:
            msg = f"Seq number mismatch: {header}"
            raise McuMgrFlashException(msg)

        if len(data) > ImageRequestHeader.size:
            dec_msg: Dict[str, Any] = cbor2.loads(data[ImageRequestHeader.size :])
        else:
            msg = f"Complete header w/o payload: {header}"
            raise McuMgrFlashException(msg)

        if "rc" not in dec_msg:
            msg = "Missing key 'rc' in response"
            raise McuMgrFlashException(msg)

        if dec_msg["rc"] != 0:
            msg = f"Error code: '{dec_msg['rc']}' response"
            raise McuMgrFlashException(msg)

        if "off" not in dec_msg:
            msg = "Missing key 'off' in response"
            raise McuMgrFlashException(msg)

        return int(dec_msg["off"])


class McuMgrFlashProtocol:
    mtu = 512
    slot = 0

    def __init__(self, port: str):
        self._port = port
        self._uart: Optional[UartComm] = UartComm(self._port)
        self.seq_id = 0

    def close(self) -> None:
        if self._uart is not None:
            self._uart.close()
            self._uart = None

    @staticmethod
    def calc_crc16_xmodem(seed: int, data: bytes) -> int:
        for byte in data:
            seed = ((seed >> 8) | (seed << 8)) & 0xFFFF
            seed ^= byte
            seed ^= ((seed & 0xFF) >> 4) & 0xFFFF
            seed ^= (seed << 12) & 0xFFFF
            seed ^= ((seed & 0xFF) << 5) & 0xFFFF

        return seed

    def get_next_seq_id(self) -> int:
        seq_id = self.seq_id
        self.seq_id = (self.seq_id + 1) % 256

        return seq_id

    def decode_response(self) -> bytes:
        assert self._uart is not None

        bytes_read = 0
        result = b""
        expected_len = 0

        while True:
            rec_data = self._uart.read(2)

            # chunk start marker expected
            if bytes_read == 0:
                if rec_data != b"\x06\x09":
                    msg = "Incorrect start marker"
                    raise McuMgrFlashException(msg)
            else:
                if rec_data != b"\x04\x14":
                    msg = "Incorrect start marker"
                    raise McuMgrFlashException(msg)

            # read until newline
            while True:
                b = self._uart.read(1)

                if b == b"\x0a":
                    break
                else:
                    result += b
                    bytes_read += 1

            dec_data = base64.b64decode(result)

            if expected_len == 0:
                length = int.from_bytes(dec_data[:2], byteorder="big")
                if length > 0:
                    expected_len = length

            # stop when done
            if len(dec_data) >= expected_len:
                break

        # decode base64
        decoded = base64.b64decode(result)

        # verify length: must be the decoded length, minus the 2 bytes to encode the length
        length = int.from_bytes(decoded[:2], byteorder="big")

        if length != len(decoded) - 2:
            msg = "Wrong chunk length"
            raise McuMgrFlashException(msg)

        # verify checksum
        data = decoded[2 : len(decoded) - 2]
        read_crc = decoded[len(decoded) - 2 :]
        calculated_crc = McuMgrFlashProtocol.calc_crc16_xmodem(0, data).to_bytes(
            2, byteorder="big"
        )

        if read_crc != calculated_crc:
            msg = "Wrong crc"
            raise McuMgrFlashException(msg)

        return data

    def reset(self) -> None:
        assert self._uart is not None

        seq_id = self.get_next_seq_id()

        req = ResetRequest(seq_id=seq_id)
        self._uart.write(req.encode())

        decoded_data = self.decode_response()

        ResetResponse.decode(decoded_data, seq_id)

    def write_image(
        self, byte_array: bytes, progress_callback: Optional[Callable[[int, bool], None]] = None
    ) -> None:
        assert self._uart is not None

        offset = 0
        byte_array_len = len(byte_array)

        while True:
            offset_start = 0
            try_length = McuMgrFlashProtocol.mtu
            seq_id = self.get_next_seq_id()

            while True:
                image_num = McuMgrFlashProtocol.slot

                if offset + try_length > byte_array_len:
                    try_length = byte_array_len - offset

                chunk = byte_array[offset : offset + try_length]

                # Include SHA and total data length only in first ImageUploadRequest
                if offset == 0:
                    sha = hashlib.sha256(byte_array).digest()
                    data_len = byte_array_len
                else:
                    sha = None
                    data_len = None

                req = ImageUploadRequest(
                    image_num=image_num,
                    offset=offset,
                    data_len=data_len,
                    data=chunk,
                    data_sha=sha,
                    seq_id=seq_id,
                )

                enc_data, enc_data_len = req.encode()

                # Test if encoded data is larger than MTU
                if enc_data_len > McuMgrFlashProtocol.mtu:
                    reduce = enc_data_len - McuMgrFlashProtocol.mtu
                    if reduce > try_length:
                        msg = "MTU too small"
                        raise McuMgrFlashException(msg)

                    # number of bytes to reduce is base64 encoded, calculate back the number of bytes
                    # and then reduce a bit more for base64 filling and rounding
                    try_length -= int(reduce * 3 / 4 + 3)
                    continue

                # Send all frames
                for frame in enc_data:
                    self._uart.write(frame)

                decoded_data = self.decode_response()

                offset = ImageUploadResponse.decode(decoded_data, seq_id)
                break

            if offset_start == offset:
                msg = "Wrong offset received"
                raise McuMgrFlashException(msg)

            if progress_callback is not None:
                progress_callback(int(100 * offset / byte_array_len), False)

            if offset == byte_array_len:
                if progress_callback is not None:
                    progress_callback(100, True)
                break


class UartComm:
    def __init__(self, port: str):
        self._ser = serial.Serial(
            port=port,
            baudrate=MCUMGR_DEFAULT_BAUDRATE,
            parity="N",
            rtscts=False,
            timeout=MCUMGR_UART_TIMEOUT,
        )
        self._ser.reset_input_buffer()

    def close(self) -> None:
        if self._ser is not None:
            self._ser.close()
            self._ser = None

    def write(self, data: bytes) -> None:
        self._ser.write(data)

    def read(self, length: int = 1) -> bytearray:
        return bytearray(self._ser.read(length))



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import contextlib
import time
from typing import Any, Callable, Iterator, Optional, TypeVar

import packaging.version

from acconeer.exptool import _core as core
from acconeer.exptool.app.new._enums import ConnectionState, PluginState
from acconeer.exptool.app.new._exceptions import HandledException

from ._backend_logger import BackendLogger
from ._backend_plugin import BackendPlugin
from ._message import (
    ConnectionStateMessage,
    GeneralMessage,
    Message,
    PluginStateMessage,
    TimingMessage,
)
from ._tasks import Task, get_task, get_task_names, is_task


_AnyClient = core.Client[Any, Any, Any, Any, Any]


T = TypeVar("T")
MessageHandler = Callable[[Message], None]
BackendPluginFactory = Callable[[MessageHandler, str], BackendPlugin[Any]]


class Model:
    backend_plugin: Optional[BackendPlugin[Any]]
    client: Optional[_AnyClient]

    def __init__(self, task_callback: Callable[[Message], None]) -> None:
        self.backend_plugin = None
        self.client = None
        self.task_callback = task_callback
        self._logger = BackendLogger.getLogger(__name__)

    def idle(self) -> bool:
        if self.backend_plugin is None:
            return False

        backend_plugin_class_name = type(self.backend_plugin).__name__
        with self.report_timing(f"{backend_plugin_class_name}.idle()"):
            return self.backend_plugin.idle()

    def execute_task(self, task: Task) -> None:
        (name, kwargs) = task

        builtin_task = get_task(self, name)
        if builtin_task is not None:
            with self.report_timing(f"Backend.{name}()"):
                builtin_task(**kwargs)
            return

        if self.backend_plugin is not None:
            backend_plugin_class_name = type(self.backend_plugin).__name__
            plugin_task = get_task(self.backend_plugin, name)
            if plugin_task is not None:
                with self.report_timing(f"{backend_plugin_class_name}.{name}()"):
                    plugin_task(**kwargs)
                return

        defined_tasks = get_task_names(self.backend_plugin) + get_task_names(self)
        task_list = ", ".join(defined_tasks)
        msg = f"'{name}' is not a task. Available tasks are: {task_list}."
        raise RuntimeError(msg)

    @contextlib.contextmanager
    def report_timing(self, name: str) -> Iterator[None]:
        start = time.perf_counter()
        yield
        end = time.perf_counter()
        self.task_callback(TimingMessage(name=name, start=start, end=end))

    @is_task
    def connect_client(
        self,
        client_factory: Callable[[], _AnyClient],
        get_connection_warning: Callable[[packaging.version.Version], Optional[str]],
    ) -> None:
        if self.client is not None:
            msg = "Model already has a Client. The current Client needs to be disconnected first."
            raise RuntimeError(msg)

        self.task_callback(ConnectionStateMessage(state=ConnectionState.CONNECTING))

        try:
            self.client = client_factory()
        except Exception as exc:
            self.client = None
            self.task_callback(ConnectionStateMessage(state=ConnectionState.DISCONNECTED))

            msg = f"Failed to connect:\n{exc}"
            raise HandledException(msg)

        if self.backend_plugin is not None:
            self.backend_plugin.attach_client(client=self.client)

        self.task_callback(
            ConnectionStateMessage(
                state=ConnectionState.CONNECTED,
                warning=get_connection_warning(self.client.server_info.parsed_rss_version),
            )
        )
        self.task_callback(GeneralMessage(name="server_info", data=self.client.server_info))

    @is_task
    def disconnect_client(self) -> None:
        if self.client is None:
            msg = "Backend has no client to disconnect."
            raise RuntimeError(msg)

        self.task_callback(ConnectionStateMessage(state=ConnectionState.DISCONNECTING))

        if self.backend_plugin is not None:
            self.backend_plugin.detach_client()

        try:
            self.client.close()
        except Exception as e:
            self._logger.exception(e)

        self.client = None

        self.task_callback(ConnectionStateMessage(state=ConnectionState.DISCONNECTED))

    @is_task
    def load_plugin(self, *, plugin_factory: BackendPluginFactory, key: str) -> None:
        if self.backend_plugin is not None:
            self.unload_plugin(send_callback=False)

        self.task_callback(PluginStateMessage(state=PluginState.LOADING))

        self.backend_plugin = plugin_factory(self.task_callback, key)
        self._logger.info(f"{plugin_factory.__name__} was loaded.")

        if self.client is not None and self.client.connected:
            self.backend_plugin.attach_client(client=self.client)
            self._logger.debug(f"{plugin_factory.__name__} was attached a Client")

        self.task_callback(PluginStateMessage(state=PluginState.LOADED_IDLE))

    @is_task
    def unload_plugin(self, send_callback: bool = True) -> None:
        if self.backend_plugin is None:
            return

        if send_callback:
            self.task_callback(PluginStateMessage(state=PluginState.UNLOADING))

        self.backend_plugin.teardown()
        self.backend_plugin = None
        self._logger.debug("Current BackendPlugin was torn down")

        if send_callback:
            self.task_callback(PluginStateMessage(state=PluginState.UNLOADED))



# Copyright (c) Acconeer AB, 2024-2026
# All rights reserved

from __future__ import annotations

import warnings
from typing import Any, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
from attributes_doc import attributes_doc

from acconeer.exptool import a121
from acconeer.exptool._core.class_creation.attrs import (
    attrs_ndarray_isclose,
    attrs_optional_ndarray_isclose,
)
from acconeer.exptool.a121._core.entities.configs.config_enums import IdleState, Profile
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    AlgoConfigBase,
    Controller,
    select_prf,
)
from acconeer.exptool.utils import is_power_of_2

from ._processor import Processor, ProcessorConfig, ProcessorExtraResult, ReportedDisplacement


def optional_profile_converter(profile: Optional[Profile]) -> Optional[Profile]:
    if profile is None:
        return None

    return Profile(profile)


def idle_state_converter(idle_state: IdleState) -> IdleState:
    return IdleState(idle_state)


@attributes_doc
@attrs.mutable(kw_only=True)
class ExampleAppConfig(AlgoConfigBase):
    measured_point: int = attrs.field(default=80)
    """Measured point."""

    time_series_length: int = attrs.field(default=1024)
    """Length of time series.
    This value will be overridden by the number of sweeps per frame if continuous sweep mode is not enabled."""

    lp_coeff: float = attrs.field(default=0.95)
    """Specify filter coefficient of the exponential filter for the FFT.
    A higher value means more filtering, i.e., slower response to changes."""

    threshold_margin: float = attrs.field(default=10.0)
    """Specify threshold margin (micrometer)."""

    amplitude_threshold: float = attrs.field(default=100.0)
    """Specify minimum amplitude for calculating vibration."""

    reported_displacement_mode: ReportedDisplacement = attrs.field(
        default=ReportedDisplacement.AMPLITUDE,
        converter=ReportedDisplacement,
    )
    """Selects whether to report the amplitude or peak to peak of the estimated frequency."""

    low_frequency_enhancement: bool = attrs.field(default=False)
    """Adds a loopback subsweep for phase correction to enhance low frequency detection."""

    profile: a121.Profile = attrs.field(default=a121.Profile.PROFILE_3, converter=a121.Profile)
    """Sets the profile."""

    frame_rate: Optional[float] = attrs.field(default=None)
    """Frame rate in Hz."""

    sweep_rate: float = attrs.field(default=3000)
    """Sweep rate in Hz."""

    sweeps_per_frame: int = attrs.field(default=128)
    """Number of sweeps per frame."""

    hwaas: int = attrs.field(default=16)
    """Number of HWAAS."""

    double_buffering: bool = attrs.field(default=True)
    """Enables double buffering."""

    continuous_sweep_mode: bool = attrs.field(default=True)
    """Enables continuous sweep mode."""

    inter_frame_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.READY, converter=idle_state_converter
    )
    """Sets the inter frame idle state."""

    inter_sweep_idle_state: a121.IdleState = attrs.field(
        default=a121.IdleState.READY, converter=idle_state_converter
    )
    """Sets the inter sweep idle state."""

    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if not is_power_of_2(self.time_series_length):
            validation_results.append(
                a121.ValidationWarning(
                    self,
                    "time_series_length",
                    "Should be power of 2 for efficient usage of fast fourier transform",
                )
            )

        if self.sweep_rate is None:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "sweep_rate",
                    "Must be set",
                )
            )

        if self.continuous_sweep_mode and not self.double_buffering:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "continuous_sweep_mode",
                    "Continuous sweep mode requires double buffering to be enabled",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ExampleAppResult:
    max_sweep_amplitude: float
    """Max amplitude in sweep.

    Used to determine whether an object is in front of the sensor.
    """

    lp_displacements: Optional[npt.NDArray[np.float64]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    """Array of estimated displacement (μm) per frequency."""

    lp_displacements_freqs: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    """Array of frequencies where displacement is estimated (Hz)."""

    max_displacement: Optional[float] = attrs.field(default=None)
    """Largest detected displacement (μm)."""

    max_displacement_freq: Optional[float] = attrs.field(default=None)
    """Frequency of largest detected displacement (Hz)."""

    time_series_std: Optional[float] = attrs.field(default=None)
    """Time series standard deviation."""

    processor_extra_result: ProcessorExtraResult = attrs.field()
    """Processor extra result, used for plotting only."""

    service_result: a121.Result = attrs.field()


class ExampleApp(Controller[ExampleAppConfig, ExampleAppResult]):
    def __init__(
        self,
        *,
        client: a121.Client,
        sensor_id: int,
        example_app_config: ExampleAppConfig,
    ) -> None:
        super().__init__(client=client, config=example_app_config)
        self.sensor_id = sensor_id

        self.started = False

    def start(
        self, recorder: Optional[a121.Recorder] = None, _algo_group: Optional[h5py.Group] = None
    ) -> None:
        if self.started:
            msg = "Already started"
            raise RuntimeError(msg)

        sensor_config = self._get_sensor_config(self.config)
        self.session_config = a121.SessionConfig(
            {self.sensor_id: sensor_config},
            extended=False,
        )

        metadata = self.client.setup_session(self.session_config)
        assert isinstance(metadata, a121.Metadata)

        processor_config = self._get_processor_config(self.config)

        self.processor = Processor(
            sensor_config=sensor_config,
            metadata=metadata,
            processor_config=processor_config,
            context=None,
        )

        if recorder is not None:
            if isinstance(recorder, a121.H5Recorder):
                if _algo_group is None:
                    _algo_group = recorder.require_algo_group("vibration")
                _record_algo_data(
                    _algo_group,
                    self.sensor_id,
                    self.config,
                )
            else:
                # Should never happen as we currently only have the H5Recorder
                warnings.warn("Will not save algo data")

            self.client.attach_recorder(recorder)

        self.client.start_session()

        self.started = True

    @classmethod
    def _get_sensor_config(cls, config: ExampleAppConfig) -> a121.SensorConfig:
        measure_config = a121.SubsweepConfig(
            start_point=config.measured_point,
            num_points=1,
            profile=config.profile,
            hwaas=config.hwaas,
            prf=select_prf(config.measured_point, config.profile),
            receiver_gain=10,
        )

        if config.low_frequency_enhancement:
            loopback_config = a121.SubsweepConfig(
                profile=a121.Profile.PROFILE_5,
                enable_loopback=True,
                hwaas=8,
                start_point=0,
                num_points=1,
                prf=select_prf(0, config.profile),
                receiver_gain=10,
            )

            subsweeps = [measure_config, loopback_config]
        else:
            subsweeps = [measure_config]

        sensor_config = a121.SensorConfig(
            subsweeps=subsweeps,
            sweeps_per_frame=config.sweeps_per_frame,
            sweep_rate=config.sweep_rate,
            continuous_sweep_mode=config.continuous_sweep_mode,
            double_buffering=config.double_buffering,
            inter_frame_idle_state=config.inter_frame_idle_state,
            inter_sweep_idle_state=config.inter_sweep_idle_state,
        )

        return sensor_config

    @classmethod
    def _get_processor_config(cls, config: ExampleAppConfig) -> ProcessorConfig:
        return ProcessorConfig(
            time_series_length=config.time_series_length,
            lp_coeff=config.lp_coeff,
            threshold_margin=config.threshold_margin,
            amplitude_threshold=config.amplitude_threshold,
            reported_displacement_mode=config.reported_displacement_mode,
            low_frequency_enhancement=config.low_frequency_enhancement,
        )

    def get_next(self) -> ExampleAppResult:
        if not self.started:
            msg = "Not started"
            raise RuntimeError(msg)

        result = self.client.get_next()
        assert isinstance(result, a121.Result)

        assert self.processor is not None
        processor_result = self.processor.process(result)

        return ExampleAppResult(
            max_sweep_amplitude=processor_result.max_sweep_amplitude,
            lp_displacements=processor_result.lp_displacements,
            lp_displacements_freqs=processor_result.lp_displacements_freqs,
            max_displacement=processor_result.max_displacement,
            max_displacement_freq=processor_result.max_displacement_freq,
            time_series_std=processor_result.time_series_std,
            processor_extra_result=processor_result.extra_result,
            service_result=result,
        )

    def update_config(self, config: ExampleAppConfig) -> None:
        raise NotImplementedError

    def stop(self) -> Any:
        if not self.started:
            msg = "Already stopped"
            raise RuntimeError(msg)

        self.client.stop_session()
        recorder = self.client.detach_recorder()
        if recorder is None:
            recorder_result = None
        else:
            recorder_result = recorder.close()

        self.started = False

        return recorder_result


def _record_algo_data(
    algo_group: h5py.Group,
    sensor_id: int,
    config: ExampleAppConfig,
) -> None:
    algo_group.create_dataset(
        "sensor_id",
        data=sensor_id,
        track_times=False,
    )
    _create_h5_string_dataset(algo_group, "example_app_config", config.to_json())


def _load_algo_data(algo_group: h5py.Group) -> Tuple[int, ExampleAppConfig]:
    sensor_id = algo_group["sensor_id"][()]
    config = ExampleAppConfig.from_json(algo_group["example_app_config"][()])
    return sensor_id, config



# Copyright (c) Acconeer AB, 2024
# All rights reserved

"""
This module does a best effort to warn the user of there being
multiple Qt bindings installed in the current Python environment,
prompting them to resolve the issue.
"""

import importlib.metadata


_WARNING_FMT = """\
============================================= WARNING ==============================================
Found conflicting Qt binding '{binding_module_name}'.

Exploration Tool might not function properly with it installed.

Resolve the issue by either:

- Installing Exploration Tool in a virtualenv.
  See https://packaging.python.org/en/latest/guides/installing-using-pip-and-virtual-environments/

- Uninstall '{binding_module_name}' with the command
  * pip uninstall {binding_module_name}  (Windows)
  * pip3 uninstall {binding_module_name} (Linux)
===================================================================================================
"""


try:
    import PySide6  # noqa: F401
except ImportError:
    pass  # Did not find installed PySide6 (which is OK)

for disallowed_binding in ["PyQt6", "PyQt5", "PySide2"]:
    try:
        _ = importlib.metadata.version(disallowed_binding)
    except importlib.metadata.PackageNotFoundError:
        pass
    else:
        print(_WARNING_FMT.format(binding_module_name=disallowed_binding))



# Copyright (c) Acconeer AB, 2023
# All rights reserved

"""
This module provides classes for listing- & creating sessions
according to different "session schemas" (how sessions are laid out inside a H5 file).
"""

from __future__ import annotations

import itertools
import typing as t

import h5py


class SessionSchema:
    @staticmethod
    def _session_names() -> t.Iterator[str]:
        return (f"sessions/session_{i}" for i in itertools.count(start=0))

    @staticmethod
    def _aliased_group(file: h5py.File, session_name: str) -> t.Optional[h5py.Group]:
        if session_name == "sessions/session_0":
            return file.get("session")
        else:
            return None

    @classmethod
    def session_groups_on_disk(cls, file: h5py.File) -> t.Sequence[h5py.Group]:
        """Lists sessions in a h5 file"""
        return tuple(
            itertools.takewhile(
                lambda group: group is not None,
                (
                    file.get(name, default=cls._aliased_group(file, name))
                    for name in cls._session_names()
                ),
            )
        )

    @classmethod
    def create_next_session_group(cls, file: h5py.File) -> h5py.Group:
        """Creates a new session group in the h5 file."""
        next_name = next(itertools.dropwhile(lambda name: name in file, cls._session_names()))

        if next_name == "sessions/session_0":
            file["session"] = h5py.SoftLink("/sessions/session_0")

        return file.create_group(next_name)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

import pyqtgraph as pg

import acconeer.exptool as et


class PGUpdater:
    def __init__(self, sensor_config, processing_config, session_info):
        self.sensor_config = sensor_config
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.history_length_s = processing_config.history_length_s
        self.fs = sensor_config.update_rate

    def setup(self, win):
        win.setWindowTitle("Tank level short range")

        self.plot = win.addPlot()
        self.plot.setMenuEnabled(False)
        self.plot.setMouseEnabled(x=False, y=False)
        self.plot.hideButtons()
        self.plot.showGrid(x=True, y=True)
        self.plot.setLabel("bottom", "Depth (m)")
        self.plot.setLabel("left", "Amplitude")
        self.plot.setXRange(self.depths[0], self.depths[-1])
        self.text_x_pos = (self.depths[-1] - self.depths[0]) / 2 + self.depths[0]
        self.text_y_pos = 1.0

        win.nextRow()

        # Detection history Plot
        self.hist_plot = win.addPlot(title="Detected peaks")
        self.hist_plot.setMenuEnabled(False)
        self.hist_plot.setMouseEnabled(x=False, y=False)
        self.hist_plot.hideButtons()
        self.hist_plot.showGrid(x=True, y=True)
        self.hist_plot.addLegend()
        self.hist_plot.setLabel("bottom", "Time history (s)")
        self.hist_plot.setLabel("left", "Distance (mm)")
        self.hist_plot.setXRange(-self.history_length_s, 0)
        self.hist_plot.setYRange(1000.0 * self.depths[0], 1000.0 * self.depths[-1])

        self.main_peak = self.hist_plot.plot(
            pen=None,
            symbol="o",
            symbolSize=8,
            symbolPen="k",
            symbolBrush=et.utils.color_cycler(0),
            name="Main peak",
        )

        self.smooth_data = self.plot.plot(pen=et.utils.pg_pen_cycler(1))
        self.smooth_max = et.utils.SmoothMax(self.sensor_config.update_rate)

    def set_text(self, text):
        self.guess_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:16pt;">'
            "{}</span></div>"
        )
        guess_html = self.guess_html_format.format(text)
        self.guess_text_item = pg.TextItem(
            html=guess_html,
            fill=pg.mkColor(255, 140, 0),
            anchor=(0.5, 0),
        )

        self.guess_text_item.setPos(self.text_x_pos, self.text_y_pos)
        self.plot.addItem(self.guess_text_item)
        self.guess_text_item.setVisible(True)

    def update(self, plot_data):
        smooth_data = plot_data["smooth_data"]
        best_guess = plot_data["best_guess"]

        guess_history = plot_data["guess_hist"]
        guess_hist_indexes = plot_data["guess_hist_idx"]

        guess_str = "Best guess: {:.4f}".format(best_guess)
        self.smooth_data.setData(self.depths, smooth_data)
        x_indexes = (guess_hist_indexes - plot_data["index"]) / self.fs
        self.main_peak.setData(x_indexes, 1000 * guess_history)

        self.plot.setYRange(0.0, 1.2)
        self.set_text(guess_str)



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from acconeer.exptool import a121

from ._ref_app import PresenceWakeUpConfig, PresenceZoneConfig, RefAppConfig


def get_short_range_config() -> RefAppConfig:
    return RefAppConfig(
        wake_up_mode=True,
        wake_up_config=PresenceWakeUpConfig(
            start_m=0.5,
            end_m=1.0,
            frame_rate=2,
            sweeps_per_frame=16,
            automatic_subsweeps=True,
            signal_quality=30.0,
            hwaas=16,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.5,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=1,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=5,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=1,
            num_zones_for_wake_up=1,
        ),
        nominal_config=PresenceZoneConfig(
            start_m=0.06,
            end_m=1.0,
            frame_rate=10,
            sweeps_per_frame=16,
            automatic_subsweeps=True,
            signal_quality=30.0,
            hwaas=16,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.4,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=1,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=5,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=5,
        ),
    )


def get_medium_range_config() -> RefAppConfig:
    return RefAppConfig(
        wake_up_mode=True,
        wake_up_config=PresenceWakeUpConfig(
            start_m=1.5,
            end_m=2.5,
            frame_rate=2,
            sweeps_per_frame=16,
            automatic_subsweeps=True,
            signal_quality=20.0,
            hwaas=32,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.5,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=1,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=6,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=1,
            num_zones_for_wake_up=1,
        ),
        nominal_config=PresenceZoneConfig(
            start_m=0.3,
            end_m=2.5,
            frame_rate=12,
            sweeps_per_frame=16,
            automatic_subsweeps=True,
            signal_quality=20.0,
            hwaas=32,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.3,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=1,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=6,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=6,
        ),
    )


def get_long_range_config() -> RefAppConfig:
    return RefAppConfig(
        wake_up_mode=True,
        wake_up_config=PresenceWakeUpConfig(
            start_m=3,
            end_m=5,
            frame_rate=2,
            sweeps_per_frame=32,
            automatic_subsweeps=True,
            signal_quality=10.0,
            hwaas=48,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.4,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=1,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=6,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=5,
            num_zones_for_wake_up=2,
        ),
        nominal_config=PresenceZoneConfig(
            start_m=2,
            end_m=5,
            frame_rate=12,
            sweeps_per_frame=32,
            automatic_subsweeps=True,
            signal_quality=10.0,
            hwaas=48,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.2,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=0.8,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=6,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=5,
        ),
    )


def get_ceiling_config() -> RefAppConfig:
    return RefAppConfig(
        wake_up_mode=True,
        wake_up_config=PresenceWakeUpConfig(
            start_m=2,
            end_m=3.5,
            frame_rate=4,
            sweeps_per_frame=24,
            automatic_subsweeps=True,
            signal_quality=13.0,
            hwaas=32,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.5,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.25,
            inter_enable=True,
            inter_detection_threshold=1,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=6,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=0.5,
            inter_frame_presence_timeout=3,
            num_zones=1,
            num_zones_for_wake_up=1,
        ),
        nominal_config=PresenceZoneConfig(
            start_m=1.5,
            end_m=3.5,
            frame_rate=12,
            sweeps_per_frame=18,
            automatic_subsweeps=True,
            signal_quality=13.0,
            hwaas=18,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.2,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=0.8,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=6,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=1,
        ),
    )


def get_low_power_config() -> RefAppConfig:
    return RefAppConfig(
        wake_up_mode=True,
        wake_up_config=PresenceWakeUpConfig(
            start_m=0.38,
            end_m=0.67,
            frame_rate=0.7,
            automatic_subsweeps=False,
            signal_quality=20.0,
            sweeps_per_frame=8,
            hwaas=8,
            profile=a121.Profile.PROFILE_5,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.7,
            intra_frame_time_const=0.3,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=1.2,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=5,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=0.5,
            inter_frame_presence_timeout=2,
            num_zones=1,
            num_zones_for_wake_up=1,
        ),
        nominal_config=PresenceZoneConfig(
            start_m=0.38,
            end_m=0.67,
            frame_rate=5,
            sweeps_per_frame=16,
            automatic_subsweeps=True,
            signal_quality=20.0,
            hwaas=8,
            inter_frame_idle_state=a121.IdleState.DEEP_SLEEP,
            intra_enable=True,
            intra_detection_threshold=1.4,
            intra_frame_time_const=0.15,
            intra_output_time_const=0.3,
            inter_enable=True,
            inter_detection_threshold=1,
            inter_frame_slow_cutoff=0.2,
            inter_frame_fast_cutoff=5,
            inter_frame_deviation_time_const=0.5,
            inter_output_time_const=2,
            inter_frame_presence_timeout=3,
            num_zones=1,
        ),
    )



# Copyright (c) Acconeer AB, 2022-2024
# All rights reserved

from __future__ import annotations

import abc
import json
from typing import Any, ClassVar, List, Optional, Type, Union

import attrs

from acconeer.exptool._core.class_creation.formatting import pretty_dict_line_strs


class ClientInfoCreationError(Exception):
    pass


@attrs.frozen(slots=False)
class ConnectionTypeBase(abc.ABC):
    __registry: ClassVar[List[Type[ConnectionTypeBase]]] = []

    @classmethod
    def _register(cls, subclass: Type[ConnectionTypeBase]) -> Type[ConnectionTypeBase]:
        """Registers a subclass"""
        if not issubclass(subclass, cls):
            msg = f"{subclass.__name__!r} needs to be a subclass of {cls.__name__}."
            raise TypeError(msg)
        cls.__registry.append(subclass)
        return subclass

    @classmethod
    def _get_subclasses(cls) -> List[Type[ConnectionTypeBase]]:
        return cls.__registry

    @classmethod
    @abc.abstractmethod
    def _from_open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        flow_control: bool = True,
    ) -> ClientInfo: ...

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @property
    def serial(self) -> Optional[SerialInfo]:
        return None

    @property
    def usb(self) -> Optional[USBInfo]:
        return None

    @property
    def socket(self) -> Optional[SocketInfo]:
        return None

    @property
    def mock(self) -> Optional[MockInfo]:
        return None


@ConnectionTypeBase._register
@attrs.frozen(kw_only=True, slots=False)
class SerialInfo(ConnectionTypeBase):
    port: str
    override_baudrate: Optional[int] = None
    serial_number: Optional[str] = None
    flow_control: bool = True

    @classmethod
    def _from_open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        flow_control: bool = True,
    ) -> ClientInfo:
        if serial_port is not None:
            return ClientInfo(
                serial=SerialInfo(
                    port=serial_port,
                    override_baudrate=override_baudrate,
                    flow_control=flow_control,
                )
            )

        raise ClientInfoCreationError()

    @property
    def serial(self) -> SerialInfo:
        return self


@ConnectionTypeBase._register
@attrs.frozen(kw_only=True, slots=False)
class USBInfo(ConnectionTypeBase):
    vid: Optional[int] = None
    pid: Optional[int] = None
    serial_number: Optional[str] = None

    @classmethod
    def _from_open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        flow_control: bool = True,
    ) -> ClientInfo:
        if usb_device is not None:
            if isinstance(usb_device, bool):
                if usb_device:
                    return ClientInfo(usb=USBInfo())
                msg = "usb_device=False is not valid"
                raise ValueError(msg)

            if isinstance(usb_device, str):
                return ClientInfo(usb=USBInfo(serial_number=usb_device))

        raise ClientInfoCreationError()

    @property
    def usb(self) -> USBInfo:
        return self


@ConnectionTypeBase._register
@attrs.frozen(kw_only=True, slots=False)
class SocketInfo(ConnectionTypeBase):
    ip_address: str
    tcp_port: Optional[int]

    @classmethod
    def _from_open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        flow_control: bool = True,
    ) -> ClientInfo:
        if ip_address is not None:
            return ClientInfo(socket=SocketInfo(ip_address=ip_address, tcp_port=tcp_port))

        raise ClientInfoCreationError()

    @property
    def socket(self) -> SocketInfo:
        return self


@ConnectionTypeBase._register
@attrs.frozen(kw_only=True, slots=True)
class MockInfo(ConnectionTypeBase):
    @classmethod
    def _from_open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        flow_control: bool = True,
    ) -> ClientInfo:
        if mock is not None:
            if mock:
                return ClientInfo(mock=MockInfo())

            msg = "mock=False is not valid"
            raise ValueError(msg)

        raise ClientInfoCreationError()

    @property
    def mock(self) -> MockInfo:
        return self


@attrs.frozen(kw_only=True)
class ClientInfo:
    socket: Optional[SocketInfo] = None
    serial: Optional[SerialInfo] = None
    usb: Optional[USBInfo] = None
    mock: Optional[MockInfo] = None

    @classmethod
    def _from_open(
        cls,
        ip_address: Optional[str] = None,
        tcp_port: Optional[int] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, bool]] = None,
        mock: Optional[bool] = None,
        override_baudrate: Optional[int] = None,
        flow_control: bool = True,
    ) -> ClientInfo:
        client_info = None
        for subclass in ConnectionTypeBase._get_subclasses():
            try:
                client_info = subclass._from_open(
                    ip_address=ip_address,
                    tcp_port=tcp_port,
                    serial_port=serial_port,
                    usb_device=usb_device,
                    mock=mock,
                    override_baudrate=override_baudrate,
                    flow_control=flow_control,
                )
                break
            except ClientInfoCreationError:
                pass

        if client_info is not None:
            return client_info

        # Return empyt ClientInfo if everything is None
        return ClientInfo()

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @staticmethod
    def _migrate_pre_v6_dict(d: dict[str, Any]) -> dict[str, Any]:
        ip_address = d.get("ip_address")
        serial_port = d.get("serial_port")
        usb_device = d.get("usb_device")

        return {
            "mock": {} if d.get("mock", False) else None,
            "socket": (
                None
                if ip_address is None
                else {
                    "ip_address": ip_address,
                    "tcp_port": d.get("tcp_port"),
                }
            ),
            "serial": (
                None
                if serial_port is None
                else {
                    "port": serial_port,
                    "override_baudrate": d.get("override_baudrate"),
                }
            ),
            "usb": (
                None
                if usb_device is None
                else {
                    "vid": usb_device["vid"],
                    "pid": usb_device["pid"],
                    "serial_number": usb_device["serial"],
                }
            ),
        }

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> ClientInfo:
        v6_keys = ["socket", "serial", "usb", "mock"]
        v5_keys = [
            "ip_address",
            "tcp_port",
            "serial_port",
            "usb_device",
            "mock",
            "override_baudrate",
        ]

        if not set(d.keys()).issubset(v6_keys):
            d_extra_keys = {k for k in d if k not in v5_keys}
            if d_extra_keys == set():
                d_only_v5_keys = {k: d.get(k) for k in v5_keys}
                d = cls._migrate_pre_v6_dict(d_only_v5_keys)
            else:
                msg = f"Cannot load the dict {d} into a ClientInfo."
                raise TypeError(msg)

        serial = SerialInfo(**d["serial"]) if d.get("serial") is not None else None
        usb = USBInfo(**d["usb"]) if d.get("usb") is not None else None
        socket = SocketInfo(**d["socket"]) if d.get("socket") is not None else None
        mock = MockInfo(**d["mock"]) if d.get("mock") is not None else None
        return cls(
            socket=socket,
            serial=serial,
            usb=usb,
            mock=mock,
        )

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ClientInfo:
        return cls.from_dict(json.loads(json_str))

    def __str__(self) -> str:
        return "\n".join(
            [
                f"{type(self).__name__}",
                *pretty_dict_line_strs(self.to_dict()),
            ]
        )
